import datetime
import os
from datetime import datetime as dt
from pathlib import Path

import ecl
import numpy as np
import pandas as pd
import pyarrow
import pytest
import yaml

from ecl2df import csv2ecl, ecl2csv, summary
from ecl2df.eclfiles import EclFiles
from ecl2df.summary import (
    _df2pyarrow,
    _fallback_date_roll,
    _fix_dframe_for_libecl,
    date_range,
    df,
    df2eclsum,
    resample_smry_dates,
    smry_meta,
)

try:
    import opm  # noqa

    HAVE_OPM = True
except ImportError:
    HAVE_OPM = False

TESTDIR = Path(__file__).absolute().parent
REEK = str(TESTDIR / "data/reek/eclipse/model/2_R001_REEK-0.DATA")
EIGHTCELLS = str(TESTDIR / "data/eightcells/EIGHTCELLS.DATA")


def test_df():
    """Test that dataframes are produced"""
    eclfiles = EclFiles(EIGHTCELLS)
    sumdf = summary.df(eclfiles)

    assert sumdf.index.name == "DATE"
    assert sumdf.index.dtype == "datetime64[ns]" or sumdf.index.dtype == "datetime64"

    assert not sumdf.empty
    assert sumdf.index.name == "DATE"
    assert not sumdf.columns.empty
    assert "FOPT" in sumdf.columns

    sumdf = summary.df(eclfiles, datetime=True)
    # (datetime=True is implicit when raw time reports are requested)
    assert sumdf.index.name == "DATE"
    assert sumdf.index.dtype == "datetime64[ns]" or sumdf.index.dtype == "datetime64"

    # Metadata should be attached using the attrs attribute on a Pandas
    # Dataframe (considered experimental by Pandas)
    assert "meta" in sumdf.attrs
    assert sumdf.attrs["meta"]["FOPR"]["unit"] == "SM3/DAY"


def test_df_column_keys():
    """Test that we can slice the dataframe on columns"""
    sumdf = summary.df(EclFiles(REEK), column_keys="FOPT")
    assert set(sumdf.columns) == {"FOPT"}
    assert set(sumdf.attrs["meta"].keys()) == {"FOPT"}

    fop_cols = {
        "FOPRS",
        "FOPT",
        "FOPRH",
        "FOPTH",
        "FOPRF",
        "FOPR",
        "FOPTS",
        "FOPTF",
        "FOPP",
    }
    sumdf = summary.df(EclFiles(REEK), column_keys="FOP*")
    assert set(sumdf.columns) == fop_cols
    assert set(sumdf.attrs["meta"].keys()) == fop_cols

    sumdf = summary.df(EclFiles(REEK), column_keys=["FOP*"])
    assert set(sumdf.columns) == fop_cols
    assert set(sumdf.attrs["meta"].keys()) == fop_cols

    sumdf = summary.df(EclFiles(REEK), column_keys=["FOPR", "FOPT"])
    assert set(sumdf.columns) == {"FOPT", "FOPR"}
    assert set(sumdf.attrs["meta"].keys()) == {"FOPT", "FOPR"}

    sumdf_no_columns = summary.df(EclFiles(REEK), column_keys=["BOGUS"])
    assert sumdf_no_columns.columns.empty
    assert all(sumdf_no_columns.index == sumdf.index)


def test_summary2df_dates():
    """Test that we have some API possibilities with ISO dates"""
    eclfiles = EclFiles(REEK)

    sumdf = summary.df(
        eclfiles,
        start_date=datetime.date(2002, 1, 2),
        end_date="2002-03-01",
        time_index="daily",
        datetime=True,
    )
    assert sumdf.index.name == "DATE"
    assert sumdf.index.dtype == "datetime64[ns]" or sumdf.index.dtype == "datetime64"

    assert len(sumdf) == 59
    assert str(sumdf.index.values[0])[0:10] == "2002-01-02"
    assert sumdf.index.values[0] == np.datetime64("2002-01-02")
    assert sumdf.index.values[-1] == np.datetime64("2002-03-01")

    sumdf = summary.df(eclfiles, time_index="last", datetime=True)
    assert len(sumdf) == 1
    assert sumdf.index.values[0] == np.datetime64("2003-01-02")

    # Leave this test for the datetime=False behaviour:
    sumdf = summary.df(eclfiles, time_index="first")
    assert len(sumdf) == 1
    assert str(sumdf.index.values[0]) == "2000-01-01"


@pytest.mark.integration
def test_ecl2csv_summary(tmp_path, mocker):
    """Test that the command line utility ecl2csv is installed and
    works with summary data"""
    tmpcsvfile = tmp_path / "sum.csv"
    mocker.patch(
        "sys.argv",
        [
            "ecl2csv",
            "summary",
            "-v",
            REEK,
            "-o",
            str(tmpcsvfile),
            "--start_date",
            "2002-01-02",
            "--end_date",
            "2003-01-02",
        ],
    )
    ecl2csv.main()
    disk_df = pd.read_csv(tmpcsvfile)
    assert len(disk_df) == 97  # Includes timestamps
    assert str(disk_df["DATE"].values[0]) == "2002-01-02 00:00:00"
    assert str(disk_df["DATE"].values[-1]) == "2003-01-02 00:00:00"

    tmpcsvfile = tmp_path / "sum.csv"
    mocker.patch(
        "sys.argv",
        [
            "ecl2csv",
            "summary",
            REEK,
            "-o",
            str(tmpcsvfile),
            "--time_index",
            "daily",
            "--start_date",
            "2002-01-02",
            "--end_date",
            "2003-01-02",
        ],
    )
    ecl2csv.main()
    disk_df = pd.read_csv(tmpcsvfile)
    assert len(disk_df) == 366
    # Pandas' csv export writes datetime64 as pure date
    # when there are no clock-times involved:
    assert str(disk_df["DATE"].values[0]) == "2002-01-02"
    assert str(disk_df["DATE"].values[-1]) == "2003-01-02"


def test_paramsupport(tmp_path, mocker):
    """Test that we can merge in parameters.txt

    This test code manipulates the paths in the checked out
    repository (as it involves some pointing upwards in the directory structure)
    It should not leave any extra files around, but requires certain filenames
    not to be under version control.
    """
    tmpcsvfile = tmp_path / "sum.csv"

    eclfiles = EclFiles(EIGHTCELLS)

    parameterstxt = Path(eclfiles.get_path()) / "parameters.txt"
    if parameterstxt.is_file():
        parameterstxt.unlink()
    parameterstxt.write_text("FOO 1\nBAR 3", encoding="utf-8")
    mocker.patch(
        "sys.argv", ["ecl2csv", "summary", EIGHTCELLS, "-o", str(tmpcsvfile), "-p"]
    )
    ecl2csv.main()
    disk_df = pd.read_csv(tmpcsvfile)
    assert "FOPT" in disk_df
    assert "FOO" in disk_df
    assert "BAR" in disk_df
    assert disk_df["BAR"].unique()[0] == 3
    parameterstxt.unlink()

    parametersyml = Path(eclfiles.get_path()) / "parameters.yml"
    if parametersyml.is_file():
        parametersyml.unlink()
    parametersyml.write_text(yaml.dump({"FOO": 1, "BAR": 3}), encoding="utf-8")
    mocker.patch(
        "sys.argv", ["ecl2csv", "summary", EIGHTCELLS, "-o", str(tmpcsvfile), "-p"]
    )
    ecl2csv.main()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert "FOPT" in disk_df
    assert "FOO" in disk_df
    assert len(disk_df["FOO"].unique()) == 1
    assert disk_df["FOO"].unique()[0] == 1
    assert "BAR" in disk_df
    assert len(disk_df["BAR"].unique()) == 1
    assert disk_df["BAR"].unique()[0] == 3

    # Test the merging from summary.df() explicitly:
    assert "FOO" in summary.df(eclfiles, params=True, paramfile=None)
    assert "FOO" not in summary.df(eclfiles, params=False, paramfile=None)
    assert "FOO" not in summary.df(eclfiles, params=None, paramfile=None)

    assert "FOO" in summary.df(eclfiles, params=False, paramfile=parametersyml)
    assert "FOO" in summary.df(eclfiles, params=None, paramfile=parametersyml)
    assert "FOO" in summary.df(eclfiles, params=None, paramfile="parameters.yml")

    # Non-existing relative path is a soft error:
    assert "FOO" not in summary.df(
        eclfiles, params=None, paramfile="notexisting/parameters.yml"
    )

    # Non-existing absolute path is a hard error:
    with pytest.raises(FileNotFoundError):
        summary.df(eclfiles, params=None, paramfile="/tmp/notexisting/parameters.yml")

    parametersyml.unlink()


def test_paramsupport_explicitfile(tmp_path, mocker):
    """Test explicit naming of parameters file from command line.

    This is a little bit tricky because the parameter file is assumed to be
    relative to the DATA file, not to working directory unless it is absolute."""

    tmpcsvfile = tmp_path / "smrywithrandomparams.txt"
    randomparamfile = tmp_path / "fooparams.txt"
    randomparamfile.write_text("FOO barrbarr\nCOM 1234", encoding="ascii")
    mocker.patch(
        "sys.argv",
        [
            "ecl2csv",
            "summary",
            "--verbose",
            EIGHTCELLS,
            "-o",
            str(tmpcsvfile),
            "--paramfile",
            str(randomparamfile),  # Absolute filepath
        ],
    )
    ecl2csv.main()
    assert pd.read_csv(tmpcsvfile)["FOO"].unique() == ["barrbarr"]
    assert pd.read_csv(tmpcsvfile)["COM"].unique() == [1234]

    # If we now change to tmp_path and give a relative filename to the parameter file,'
    # it will not be found:
    os.chdir(tmp_path)
    mocker.patch(
        "sys.argv",
        [
            "ecl2csv",
            "summary",
            "--verbose",
            EIGHTCELLS,
            "-o",
            "smry_noparams.csv",
            "--paramfile",
            Path(randomparamfile).name,  # A relative filepath
        ],
    )
    ecl2csv.main()
    assert "FOO" not in pd.read_csv("smry_noparams.csv")


def test_main_subparser(tmp_path, mocker):
    """Test command line interface with output to both CSV and arrow/feather."""
    tmpcsvfile = tmp_path / "sum.csv"
    mocker.patch("sys.argv", ["ecl2csv", "summary", EIGHTCELLS, "-o", str(tmpcsvfile)])
    ecl2csv.main()

    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert not disk_df.empty
    assert "FOPT" in disk_df

    # Test arrow output format:
    tmparrowfile = tmp_path / "sum.arrow"
    mocker.patch(
        "sys.argv",
        ["ecl2csv", "summary", "--arrow", EIGHTCELLS, "-o", str(tmparrowfile)],
    )
    ecl2csv.main()
    assert Path(tmpcsvfile).is_file()
    disk_arraydf = pyarrow.feather.read_table(tmparrowfile).to_pandas()
    assert "FOPT" in disk_arraydf

    # Alternative and equivalent command line syntax for arrow output:
    tmparrowfile_alt = tmp_path / "sum2.arrow"
    mocker.patch(
        "sys.argv", ["ecl2arrow", "summary", EIGHTCELLS, "-o", str(tmparrowfile_alt)]
    )
    ecl2csv.main()
    pd.testing.assert_frame_equal(
        disk_arraydf, pyarrow.feather.read_table(str(tmparrowfile_alt)).to_pandas()
    )

    # Not possible (yet?) to write arrow to stdout:
    mocker.patch("sys.argv", ["ecl2arrow", "summary", EIGHTCELLS, "-o", "-"])
    with pytest.raises(SystemExit):
        ecl2csv.main()


def test_datenormalization():
    """Test normalization of dates, where
    dates can be ensured to be on dategrid boundaries"""
    # realization-0 here has its last summary date at 2003-01-02
    eclfiles = EclFiles(REEK)
    daily = summary.df(eclfiles, column_keys="FOPT", time_index="daily", datetime=True)
    assert str(daily.index[-1])[0:10] == "2003-01-02"
    monthly = summary.df(
        eclfiles, column_keys="FOPT", time_index="monthly", datetime=True
    )
    assert str(monthly.index[-1])[0:10] == "2003-02-01"
    yearly = summary.df(
        eclfiles, column_keys="FOPT", time_index="yearly", datetime=True
    )
    assert str(yearly.index[-1])[0:10] == "2004-01-01"


def test_extrapolation():
    """Summary data should be possible to extrapolate into
    the future, rates should be zero, cumulatives should be constant"""
    eclfiles = EclFiles(EIGHTCELLS)
    lastfopt = summary.df(
        eclfiles, column_keys="FOPT", time_index="last", datetime=True
    )["FOPT"].values[0]
    answer = pd.DataFrame(
        # This is the maximal date for datetime64[ns]
        index=[np.datetime64("2262-04-11")],
        columns=["FOPT", "FOPR"],
        data=[[lastfopt, 0.0]],
    ).rename_axis("DATE")

    pd.testing.assert_frame_equal(
        summary.df(
            eclfiles,
            column_keys=["FOPT", "FOPR"],
            time_index="2262-04-11",
            datetime=True,
        ),
        answer,
    )
    pd.testing.assert_frame_equal(
        summary.df(
            eclfiles,
            column_keys=["FOPT", "FOPR"],
            time_index=[datetime.date(2262, 4, 11)],
            # NB: df() does not support datetime64 for time_index
            datetime=True,
        ),
        answer,
    )

    # Pandas does not support DatetimeIndex beyound 2262:
    with pytest.raises(pd.errors.OutOfBoundsDatetime):
        summary.df(
            eclfiles,
            column_keys=["FOPT"],
            time_index=[datetime.date(2300, 1, 1)],
            datetime=True,
        )

    # But without datetime, we can get it extrapolated by libecl:
    assert summary.df(
        eclfiles, column_keys=["FOPT"], time_index=[datetime.date(2300, 1, 1)]
    )["FOPT"].values == [lastfopt]


def test_foreseeable_future(tmp_path):
    """The foreseeable future in reservoir simulation is "defined" as 500 years.

    Check that we support summary files with this timespan"""
    os.chdir(tmp_path)
    src_dframe = pd.DataFrame(
        [
            {"DATE": "2000-01-01", "FPR": 200},
            {"DATE": "2500-01-01", "FPR": 180},
        ]
    )
    eclsum = df2eclsum(src_dframe, casename="PLUGABANDON")

    dframe = summary.df(eclsum)
    assert (
        dframe.index
        == [
            dt(2000, 1, 1),
            # This discrepancy is due to seconds as a 32-bit float
            # having an accuracy limit (roundoff-error)
            # https://github.com/equinor/ecl/issues/803
            dt(2499, 12, 31, 23, 55, 44),
        ]
    ).all()

    # Try with time interpolation involved:
    dframe = summary.df(eclsum, time_index="yearly")
    assert len(dframe) == 501
    assert dframe.index.max() == datetime.date(year=2500, month=1, day=1)

    # Try with one-year timesteps:
    src_dframe = pd.DataFrame(
        {
            "DATE": pd.date_range("2000-01-01", "2069-01-01", freq="YS"),
            "FPR": range(70),
        }
    )
    eclsum = df2eclsum(src_dframe, casename="PLUGABANDON")
    dframe = summary.df(eclsum)
    # Still buggy:
    assert dframe.index[-1] == dt(2068, 12, 31, 23, 57, 52)

    # Try with one-year timesteps, starting late:
    src_dframe = pd.DataFrame(
        {
            "DATE": [datetime.date(2400 + year, 1, 1) for year in range(69)],
            "FPR": range(69),
        }
    )
    eclsum = df2eclsum(src_dframe, casename="PLUGABANDON")
    dframe = summary.df(eclsum)
    # Works fine when stepping only 68 years:
    assert dframe.index[-1] == dt(2468, 1, 1, 0, 0, 0)


@pytest.mark.parametrize(
    "rollme, direction, freq, expected",
    [
        (
            dt(3000, 1, 1),
            "forward",
            "yearly",
            dt(3000, 1, 1),
        ),
        (
            dt(3000, 1, 1),
            "forward",
            "monthly",
            dt(3000, 1, 1),
        ),
        (
            dt(3000, 1, 2),
            "forward",
            "yearly",
            dt(3001, 1, 1),
        ),
        (
            dt(3000, 1, 2),
            "forward",
            "monthly",
            dt(3000, 2, 1),
        ),
        (
            dt(3000, 1, 1),
            "back",
            "yearly",
            dt(3000, 1, 1),
        ),
        (
            dt(3000, 1, 1),
            "back",
            "monthly",
            dt(3000, 1, 1),
        ),
        (
            dt(3000, 12, 31),
            "back",
            "yearly",
            dt(3000, 1, 1),
        ),
        (
            dt(3000, 2, 2),
            "back",
            "monthly",
            dt(3000, 2, 1),
        ),
        pytest.param(
            dt(3000, 2, 2),
            "forward",
            "daily",
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        pytest.param(
            dt(3000, 2, 2),
            "upwards",
            "yearly",
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
    ],
)
def test_fallback_date_roll(rollme, direction, freq, expected):
    assert _fallback_date_roll(rollme, direction, freq) == expected


@pytest.mark.parametrize(
    "start, end, freq, expected",
    [
        (
            dt(3000, 1, 1),
            dt(3002, 1, 1),
            "yearly",
            [
                dt(3000, 1, 1),
                dt(3001, 1, 1),
                dt(3002, 1, 1),
            ],
        ),
        (
            dt(2999, 11, 1),
            dt(3000, 2, 1),
            "monthly",
            [
                dt(2999, 11, 1),
                dt(2999, 12, 1),
                dt(3000, 1, 1),
                dt(3000, 2, 1),
            ],
        ),
        pytest.param(
            dt(3000, 1, 1),
            dt(3000, 2, 1),
            "weekly",
            None,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        (
            # Crossing the problematic time boundary:
            dt(2260, 1, 1),
            dt(2263, 1, 1),
            "yearly",
            [
                dt(2260, 1, 1),
                dt(2261, 1, 1),
                dt(2262, 1, 1),
                dt(2263, 1, 1),
            ],
        ),
        (
            dt(3000, 1, 1),
            dt(3000, 1, 1),
            "yearly",
            [
                dt(3000, 1, 1),
            ],
        ),
        (
            dt(2000, 1, 1),
            dt(2000, 1, 1),
            "yearly",
            [
                dt(2000, 1, 1),
            ],
        ),
        (
            dt(2000, 1, 1),
            dt(1000, 1, 1),
            "yearly",
            [],
        ),
        (
            dt(3000, 1, 1),
            dt(2000, 1, 1),
            "yearly",
            [],
        ),
        (
            dt(2300, 5, 6),
            dt(2302, 3, 1),
            "yearly",
            [
                dt(2300, 5, 6),
                dt(2301, 1, 1),
                dt(2302, 1, 1),
                dt(2302, 3, 1),
            ],
        ),
        (
            dt(2304, 5, 6),
            dt(2302, 3, 1),
            "yearly",
            [],
        ),
        (
            dt(2302, 3, 1),
            dt(2302, 3, 1),
            "yearly",
            [dt(2302, 3, 1)],
        ),
    ],
)
def test_date_range(start, end, freq, expected):
    # When dates are beyond year 2262, the function _fallback_date_range() is triggered.
    assert date_range(start, end, freq) == expected


def test_resample_smry_dates():
    """Test resampling of summary dates"""

    eclfiles = EclFiles(REEK)

    ecldates = eclfiles.get_eclsum().dates

    assert isinstance(resample_smry_dates(ecldates), list)
    assert isinstance(resample_smry_dates(ecldates, freq="last"), list)
    assert isinstance(resample_smry_dates(ecldates, freq="last")[0], datetime.date)

    monthly = resample_smry_dates(ecldates, freq="monthly")
    assert monthly[-1] > monthly[0]  # end date is later than start
    assert len(resample_smry_dates(ecldates, freq="yearly")) == 5
    assert len(monthly) == 38
    assert len(resample_smry_dates(ecldates, freq="daily")) == 1098

    weekly = resample_smry_dates(ecldates, freq="weekly")
    assert len(weekly) == 159

    assert resample_smry_dates(ecldates, freq="2001-04-05") == [dt(2001, 4, 5, 0, 0)]
    assert resample_smry_dates(ecldates, freq=datetime.date(2001, 4, 5)) == [
        datetime.date(2001, 4, 5)
    ]
    assert resample_smry_dates(ecldates, freq=dt(2001, 4, 5, 0, 0)) == [
        dt(2001, 4, 5, 0, 0)
    ]
    # start and end should be included:
    assert (
        len(
            resample_smry_dates(
                ecldates, start_date="2000-06-05", end_date="2000-06-07", freq="daily"
            )
        )
        == 3
    )
    # No month boundary between start and end, but we
    # should have the starts and ends included
    assert (
        len(
            resample_smry_dates(
                ecldates, start_date="2000-06-05", end_date="2000-06-07", freq="monthly"
            )
        )
        == 2
    )
    # Date normalization should be overriden here:
    assert (
        len(
            resample_smry_dates(
                ecldates,
                start_date="2000-06-05",
                end_date="2000-06-07",
                freq="monthly",
                normalize=True,
            )
        )
        == 2
    )
    # Start_date and end_date at the same date should work
    assert (
        len(
            resample_smry_dates(
                ecldates, start_date="2000-01-01", end_date="2000-01-01"
            )
        )
        == 1
    )
    assert (
        len(
            resample_smry_dates(
                ecldates, start_date="2000-01-01", end_date="2000-01-01", normalize=True
            )
        )
        == 1
    )

    # Check that we can go way outside the smry daterange:
    assert (
        len(
            resample_smry_dates(
                ecldates, start_date="1978-01-01", end_date="2030-01-01", freq="yearly"
            )
        )
        == 53
    )
    assert (
        len(
            resample_smry_dates(
                ecldates,
                start_date="1978-01-01",
                end_date="2030-01-01",
                freq="yearly",
                normalize=True,
            )
        )
        == 53
    )

    assert (
        len(
            resample_smry_dates(
                ecldates,
                start_date="2000-06-05",
                end_date="2000-06-07",
                freq="raw",
                normalize=True,
            )
        )
        == 2
    )
    assert (
        len(
            resample_smry_dates(
                ecldates,
                start_date="2000-06-05",
                end_date="2000-06-07",
                freq="raw",
                normalize=False,
            )
        )
        == 2
    )

    # Check that we can be outside the nanosecond timestamp limitation in Pandas:
    assert (
        len(
            resample_smry_dates(
                ecldates,
                start_date="2000-06-05",
                end_date="2300-06-07",
                freq="yearly",
                normalize=True,
            )
        )
        == 2 + 300  # boundary dates + 2001-01-01 to 2300-01-01
    )

    # Verify boundary date bug up to and including ecl2df v0.13.2
    assert resample_smry_dates(
        ecldates,
        start_date="2300-06-05",
        end_date="2301-06-07",
        freq="yearly",
        normalize=False,
    ) == [dt(2300, 6, 5).date(), dt(2301, 1, 1).date(), dt(2301, 6, 7).date()]
    # Normalization should not change anything when dates are explicit:
    assert resample_smry_dates(
        ecldates,
        start_date="2300-06-05",
        end_date="2301-06-07",
        freq="yearly",
        normalize=True,
    ) == [dt(2300, 6, 5).date(), dt(2301, 1, 1).date(), dt(2301, 6, 7).date()]


def test_smry_meta():
    """Test obtaining metadata dictionary for summary vectors from an EclSum object"""
    meta = smry_meta(EclFiles(REEK))

    assert isinstance(meta, dict)
    assert "FOPT" in meta
    assert "FOPTH" in meta
    assert meta["FOPT"]["unit"] == "SM3"
    assert meta["FOPR"]["unit"] == "SM3/DAY"
    assert meta["FOPT"]["is_total"]
    assert not meta["FOPR"]["is_total"]
    assert not meta["FOPT"]["is_rate"]
    assert meta["FOPR"]["is_rate"]
    assert not meta["FOPT"]["is_historical"]
    assert meta["FOPTH"]["is_historical"]

    assert meta["WOPR:OP_1"]["wgname"] == "OP_1"
    assert meta["WOPR:OP_1"]["keyword"] == "WOPR"
    if "wgname" in meta["FOPT"]:
        # Not enforced yet to have None fields actually included
        assert meta["FOPT"]["wgname"] is None

    # Can create dataframes like this:
    meta_df = pd.DataFrame.from_dict(meta, orient="index")
    hist_keys = meta_df[meta_df["is_historical"]].index
    assert all([key.split(":")[0].endswith("H") for key in hist_keys])


def test_smry_meta_synthetic():
    """What does meta look like when we start from a synthetic summary?

    ecl2df currently does not try to set the units to anything when
    making synthetic summary.
    """
    dframe = pd.DataFrame(
        [
            {"DATE": np.datetime64("2016-01-01"), "FOPT": 1000, "FOPR": 100},
        ]
    ).set_index("DATE")
    synt_meta = smry_meta(df2eclsum(dframe))

    # Dummy unit provided by EclSum:
    assert synt_meta["FOPT"]["unit"] == "UNIT"


@pytest.mark.parametrize(
    "dframe, expected_dframe",
    [
        # # # # # # # # # # # # # # # # # # # # # # # #
        (pd.DataFrame(), pd.DataFrame()),
        # # # # # # # # # # # # # # # # # # # # # # # #
        (
            pd.DataFrame([{"DATE": "2516-01-01", "FOPT": 1}]),
            pd.DataFrame([{"FOPT": 1}], index=[dt(2516, 1, 1)]).rename_axis("DATE"),
        ),
        # # # # # # # # # # # # # # # # # # # # # # # #
        (
            pd.DataFrame([{"DATE": np.datetime64("2016-01-01"), "FOPT": 1}]),
            pd.DataFrame([{"FOPT": 1}], index=[dt(2016, 1, 1)]).rename_axis("DATE"),
        ),
        # # # # # # # # # # # # # # # # # # # # # # # #
        (
            pd.DataFrame(
                [{"DATE": pd.Timestamp(datetime.date(2016, 1, 1)), "FOPT": 1}]
            ),
            pd.DataFrame([{"FOPT": 1}], index=[dt(2016, 1, 1)]).rename_axis("DATE"),
        ),
        # # # # # # # # # # # # # # # # # # # # # # # #
        (
            pd.DataFrame(
                [
                    {
                        "DATE": dt(2016, 1, 1, 12, 34, 56),
                        "FOPT": 1,
                    }
                ]
            ),
            pd.DataFrame([{"FOPT": 1}], index=[dt(2016, 1, 1, 12, 34, 56)]).rename_axis(
                "DATE"
            ),
        ),
        # # # # # # # # # # # # # # # # # # # # # # # #
        (
            pd.DataFrame([{"DATE": "2016-01-01", "FOPT": 1}], index=[2]),
            pd.DataFrame(
                [{"FOPT": 1}], index=[pd.to_datetime("2016-01-01")]
            ).rename_axis("DATE"),
        ),
        # # # # # # # # # # # # # # # # # # # # # # # #
        (
            pd.DataFrame([{"DATE": datetime.date(2017, 1, 1), "FOPT": 1}]),
            pd.DataFrame([{"FOPT": 1}], index=[dt(2017, 1, 1)]).rename_axis("DATE"),
        ),
        # # # # # # # # # # # # # # # # # # # # # # # #
        (
            # Dates that go beyond Pandas' datetime64[ns] limits:
            pd.DataFrame([{"DATE": datetime.date(2517, 1, 1), "FOPT": 1}]),
            pd.DataFrame([{"FOPT": 1}], index=[dt(2517, 1, 1)]).rename_axis("DATE"),
        ),
        # # # # # # # # # # # # # # # # # # # # # # # #
        (
            pd.DataFrame(
                [{"DATE": pd.to_datetime(datetime.date(2016, 1, 1)), "FOPT": 1}]
            ),
            pd.DataFrame(
                [{"FOPT": 1}], index=[pd.to_datetime("2016-01-01")]
            ).rename_axis("DATE"),
        ),
        # # # # # # # # # # # # # # # # # # # # # # # #
        (
            pd.DataFrame([{"DATE": "2016-01-01", "RGIP:1": 1}]),
            pd.DataFrame(
                [{"RGIP:1": 1}], index=[pd.to_datetime("2016-01-01")]
            ).rename_axis("DATE"),
        ),
        # # # # # # # # # # # # # # # # # # # # # # # #
        (
            pd.DataFrame([{"DATE": "2016-01-01", "BWPR:1": 1}]),
            pd.DataFrame([], index=[pd.to_datetime("2016-01-01")]).rename_axis("DATE"),
        ),
        # # # # # # # # # # # # # # # # # # # # # # # #
        (
            pd.DataFrame([{"DATE": "2016-01-01", "BWPR:100,100,100": 1}]),
            pd.DataFrame([], index=[pd.to_datetime("2016-01-01")]).rename_axis("DATE"),
        ),
        # # # # # # # # # # # # # # # # # # # # # # # #
        (
            pd.DataFrame([{"DATE": "2016-01-01", "LBPR_X:100,100,100": 1}]),
            pd.DataFrame([], index=[pd.to_datetime("2016-01-01")]).rename_axis("DATE"),
        ),
    ],
)
def test_fix_dframe_for_libecl(dframe, expected_dframe):
    """Test the dataframe preprocessor/validator for df2eclsum works"""
    pd.testing.assert_frame_equal(
        _fix_dframe_for_libecl(dframe), expected_dframe, check_index_type=False
    )


@pytest.mark.parametrize(
    "dframe",
    [
        (pd.DataFrame()),
        (pd.DataFrame([{"DATE": "2016-01-01", "FOPT": 1000}])),
        (pd.DataFrame([{"DATE": "2016-01-01", "FOPT": 1000, "FOPR": 100}])),
        (pd.DataFrame([{"DATE": "3016-01-01", "FOPT": 1000}])),
        # # # # # # # # # # # # # # # # # # # # # # # #
        (
            pd.DataFrame(
                [
                    {"DATE": "2016-01-01", "FOPT": 1000, "FOPR": 100},
                    {"DATE": "2017-01-01", "FOPT": 1000, "FOPR": 100},
                ]
            )
        ),
        # # # # # # # # # # # # # # # # # # # # # # # #
        (
            pd.DataFrame(
                [
                    {"DATE": "2016-01-01", "BPR:1,1,1": 100},
                ]
            )
        ),
        # # # # # # # # # # # # # # # # # # # # # # # #
        (
            pd.DataFrame(
                [
                    {"DATE": "2016-01-01", "RWIP:1": 100},
                ]
            )
        ),
        # # # # # # # # # # # # # # # # # # # # # # # #
        (
            pd.DataFrame(
                [
                    {"DATE": "2016-01-01", "BPR:11111": 100},
                ]
            )
        ),
        # # # # # # # # # # # # # # # # # # # # # # # #
        (
            pd.DataFrame(
                [
                    # This is not allowed in Eclipse, but works here.
                    {"DATE": "2016-01-01", "FOOBAR": 100},
                ]
            )
        ),
        # # # # # # # # # # # # # # # # # # # # # # # #
        (
            pd.DataFrame(
                [
                    # This is not allowed in Eclipse, but works here.
                    {"DATE": "2016-01-01", "FOOBARCOMBARCOM": 100},
                ]
            )
        ),
        # # # # # # # # # # # # # # # # # # # # # # # #
        (
            pd.DataFrame(
                [
                    # This is not allowed in Eclipse, but works here.
                    {"DATE": "2016-01-01", "foobarcombarcom": 100},
                ]
            )
        ),
    ],
)
def test_df2eclsum(dframe):
    """Test that a dataframe can be converted to an EclSum object, and then read
    back again"""

    # Massage the dframe first so we can assert on equivalence after.
    dframe = _fix_dframe_for_libecl(dframe)

    eclsum = df2eclsum(dframe)
    if dframe.empty:
        assert eclsum is None
        return

    dframe_roundtrip = df(eclsum)
    pd.testing.assert_frame_equal(
        dframe.sort_index(axis=1),
        dframe_roundtrip.sort_index(axis=1),
        check_dtype=False,
    )


def test_df2eclsum_datetimeindex():
    """Test that providing a dataframe with a datetimeindex also works"""
    dframe = pd.DataFrame(
        [
            {"DATE": "2016-01-01", "FOPT": 1000, "FOPR": 100},
        ]
    )
    dframe["DATE"] = pd.to_datetime(dframe["DATE"])
    dframe.set_index("DATE")

    roundtrip = df(df2eclsum(dframe))
    assert isinstance(roundtrip.index, pd.DatetimeIndex)
    assert roundtrip["FOPR"].values == [100]
    assert roundtrip["FOPT"].values == [1000]


def test_duplicated_summary_vectors(caplog):
    """EclSum files on disk may contain repeated vectors
    if the user has inserted a vector name twice in the
    SUMMARY section

    ecl2df.summary.df() should deduplicate this, and give a warning.
    """

    # ecl2df.df2eclsum() is not able to mock such a UNSMRY file.
    dupe_datafile = (
        TESTDIR
        / "data"
        / "eightcells"
        / "eightcells_duplicated_summary_vector"
        / "EIGHTCELLS_DUPES.DATA"
    )
    assert "SUMMARY\nFOPR\nFOPR" in dupe_datafile.read_text()
    deduplicated_dframe = df(EclFiles(dupe_datafile))
    assert (deduplicated_dframe.columns == ["YEARS", "FOPR"]).all()
    assert "Duplicated columns detected" in caplog.text


def test_df2pyarrow_ints():
    """Test a dummy integer table converted into PyArrow"""
    dframe = pd.DataFrame(columns=["FOO", "BAR"], data=[[1, 2], [3, 4]]).astype("int32")
    pyat_df = _df2pyarrow(dframe).to_pandas()

    pd.testing.assert_frame_equal(dframe, pyat_df[["FOO", "BAR"]])

    # Millisecond datetimes:
    assert (
        pyat_df["DATE"].to_numpy()
        == [
            np.datetime64("1970-01-01T00:00:00.000000000"),
            np.datetime64("1970-01-01T00:00:00.001000000"),  # Milliseconds
        ]
    ).all()


def test_df2pyarrow_mix_int_float():
    """Test that mixed integer and float columns are conserved"""
    dframe = pd.DataFrame(columns=["FOO", "BAR"], data=[[1, 2], [3, 4]]).astype("int32")
    dframe["BAR"] *= 1.1  # Make it into a float type.
    pyat_df = _df2pyarrow(dframe).to_pandas()

    # For the comparison:
    dframe["BAR"] = dframe["BAR"].astype("float32")

    pd.testing.assert_frame_equal(dframe, pyat_df[["FOO", "BAR"]])


def test_df2pyarrow_500years():
    """Summary files can have DATE columns with timespans outside the
    Pandas dataframe nanosecond limitation. This should not present
    a problem to the PyArrow conversion"""
    dateindex = [dt(1000, 1, 1, 0, 0, 0), dt(3000, 1, 1, 0, 0, 0)]
    dframe = pd.DataFrame(
        columns=["FOO", "BAR"], index=dateindex, data=[[1, 2], [3, 4]]
    ).astype("int32")

    # The index name should be ignored:
    dframe.index.name = "BOGUS"
    pyat = _df2pyarrow(dframe)

    with pytest.raises(pyarrow.lib.ArrowInvalid):
        # We cannot convert this back to Pandas, since it will bail on failing
        # to use nanosecond timestamps in the dataframe object for these dates.
        # This is maybe a PyArrow bug/limitation that we must be aware of.
        pyat.to_pandas()

    assert (
        np.array(pyat.column(0))
        == [
            np.datetime64("1000-01-01T00:00:00.000000000"),
            np.datetime64("3000-01-01T00:00:00.000000000"),
        ]
    ).all()


def test_df2pyarrow_meta():
    """Test that metadata in summary dataframes dframe.attrs are passed on to
    pyarrow tables"""
    dframe = pd.DataFrame(columns=["FOO", "BAR"], data=[[1, 2], [3, 4]]).astype("int32")
    dframe.attrs["meta"] = {
        "FOO": {"unit": "barf", "is_interesting": False},
        "ignoreme": "ignored",
    }
    pyat = _df2pyarrow(dframe)
    assert pyat.select(["FOO"]).schema[0].metadata == {
        b"unit": b"barf",
        b"is_interesting": b"False",
    }
    assert pyat.select(["BAR"]).schema[0].metadata == {}

    assert "is_interesting" in pyat.schema.to_string()
    assert "ignored" not in pyat.schema.to_string()


def test_df2pyarrow_strings():
    """Check that dataframes can have string columns passing through PyArrow"""
    dframe = pd.DataFrame(columns=["FOO", "BAR"], data=[["hei", "hopp"]])
    pyat_df = _df2pyarrow(dframe).to_pandas()
    pd.testing.assert_frame_equal(dframe, pyat_df[["FOO", "BAR"]])


@pytest.mark.skipif(not HAVE_OPM, reason="Test requires OPM")
def test_ecl2df_errors(tmp_path):
    """Test error handling on bogus/corrupted summary files"""
    os.chdir(tmp_path)
    Path("FOO.UNSMRY").write_bytes(os.urandom(100))
    Path("FOO.SMSPEC").write_bytes(os.urandom(100))
    with pytest.raises(OSError, match="Failed to create summary instance"):
        # This is how libecl reacts to bogus binary data
        ecl.summary.EclSum("FOO.UNSMRY")

    # But EclFiles should be more tolerant, as it should be possible
    # to extract other data if SMRY is corrupted
    Path("FOO.DATA").write_text("RUNSPEC")
    assert str(EclFiles("FOO").get_ecldeck()).strip() == "RUNSPEC"
    with pytest.raises(OSError):
        EclFiles("FOO").get_eclsum()

    # Getting a dataframe from bogus data should give empty data:
    assert df(EclFiles("FOO")).empty


def test_df2eclsum_errors():
    """Test various error conditions, checking that the correct error message
    is emitted"""
    dframe = pd.DataFrame(
        [
            {"DATE": "2016-01-01", "FOPT": 1000, "FOPR": 100},
        ]
    )
    with pytest.raises(ValueError, match="casename foobar must be UPPER CASE"):
        df2eclsum(dframe, casename="foobar")
    with pytest.raises(ValueError, match="Do not use dots in casename"):
        df2eclsum(dframe, casename="FOOBAR.UNSMRY")  # .UNSMRY should not be included

    # No date included:
    with pytest.raises(ValueError, match="dataframe must have a datetime index"):
        df2eclsum(pd.DataFrame([{"FOPT": 1000}]))


@pytest.mark.integration
def test_csv2ecl_summary(tmp_path, mocker):
    """Check that we can call df2eclsum through the csv2ecl command line
    utility"""
    dframe = pd.DataFrame(
        [
            {"DATE": "2016-01-01", "FOPT": 1000, "FOPR": 100},
            {"DATE": "2017-01-01", "FOPT": 1000, "FOPR": 100},
        ]
    )
    os.chdir(tmp_path)
    dframe.to_csv("summary.csv")
    mocker.patch(
        "sys.argv",
        [
            "csv2ecl",
            "summary",
            "-v",
            "summary.csv",
            "--output",
            "SYNTHETIC",
        ],
    )
    csv2ecl.main()
    assert Path("SYNTHETIC.UNSMRY").is_file()
    assert Path("SYNTHETIC.SMSPEC").is_file()

    # Check that we can write to a subdirectory
    Path("foo").mkdir()
    mocker.patch(
        "sys.argv",
        [
            "csv2ecl",
            "summary",
            "--debug",
            "summary.csv",
            "--output",
            str(Path("foo") / Path("SYNTHETIC")),
        ],
    )
    csv2ecl.main()
    assert ("foo" / Path("SYNTHETIC.UNSMRY")).is_file()
    assert ("foo" / Path("SYNTHETIC.SMSPEC")).is_file()



"""Test module for nnc2df"""

import io
import os
import subprocess
from pathlib import Path

import pandas as pd
import pytest

from ecl2df import ecl2csv, faults, nnc, trans
from ecl2df.eclfiles import EclFiles

try:
    import opm  # noqa

    HAVE_OPM = True
except ImportError:
    HAVE_OPM = False


TESTDIR = Path(__file__).absolute().parent
REEK = str(TESTDIR / "data/reek/eclipse/model/2_R001_REEK-0.DATA")
EIGHTCELLS = str(TESTDIR / "data/eightcells/EIGHTCELLS.DATA")


def test_nnc2df():
    """Test that dataframes are produced"""
    eclfiles = EclFiles(REEK)
    nncdf = nnc.df(eclfiles)

    assert not nncdf.empty
    assert "I1" in nncdf
    assert "J1" in nncdf
    assert "K1" in nncdf
    assert "I2" in nncdf
    assert "J2" in nncdf
    assert "K2" in nncdf
    assert "TRAN" in nncdf

    prelen = len(nncdf)
    nncdf = nnc.filter_vertical(nncdf)
    assert (nncdf["I1"] == nncdf["I2"]).all()
    assert (nncdf["J1"] == nncdf["J2"]).all()
    assert len(nncdf) < prelen


def test_no_nnc():
    """Test nnc on an Eclipse case with no NNCs"""
    eclfiles = EclFiles(EIGHTCELLS)
    assert nnc.df(eclfiles).empty


def test_nnc2df_coords():
    """Test that we are able to add coordinates"""
    eclfiles = EclFiles(REEK)
    gnncdf = nnc.df(eclfiles, coords=True)
    assert not gnncdf.empty
    assert "X" in gnncdf
    assert "Y" in gnncdf
    assert "Z" in gnncdf


@pytest.mark.skipif(not HAVE_OPM, reason="Requires OPM")
def test_nnc2df_faultnames():
    """Add faultnames from FAULTS keyword to connections"""
    eclfiles = EclFiles(REEK)
    nncdf = nnc.df(eclfiles)
    faultsdf = faults.df(eclfiles.get_ecldeck())

    merged = pd.merge(
        nncdf,
        faultsdf,
        how="left",
        left_on=["I1", "J1", "K1"],
        right_on=["I", "J", "K"],
    )
    merged = pd.merge(
        merged,
        faultsdf,
        how="left",
        left_on=["I2", "J2", "K2"],
        right_on=["I", "J", "K"],
    )
    # Fix columnnames so that we don't get FACE_x and FACE_y etc.
    # Remove I_x, J_x, K_x (and _y) which is not needed


def test_df2ecl_editnnc(tmp_path):
    """Test generation of EDITNNC keyword"""
    eclfiles = EclFiles(REEK)
    nncdf = nnc.df(eclfiles)
    os.chdir(tmp_path)

    nncdf["TRANM"] = 2
    editnnc = nnc.df2ecl_editnnc(nncdf, filename="editnnc.inc")
    editnnc_fromfile = "".join(open("editnnc.inc").readlines())
    assert editnnc == editnnc_fromfile
    assert "EDITNNC" in editnnc
    assert editnnc.count("/") == len(nncdf) + 1
    assert "avg multiplier" in editnnc

    # Fails when columns are missing
    with pytest.raises((KeyError, ValueError)):
        nnc.df2ecl_editnnc(nncdf[["I1", "I2"]])

    editnnc = nnc.df2ecl_editnnc(nncdf, nocomments=True)
    assert "avg multiplier" not in editnnc

    # Test compatibility with trans module:
    trans_df = trans.df(eclfiles, addnnc=True)
    editnnc = nnc.df2ecl_editnnc(trans_df.assign(TRANM=0.3))
    assert "avg multiplier 0.3" in editnnc or "avg multiplier 0.29999" in editnnc

    print(nnc.df2ecl_editnnc(nnc.df(eclfiles).head(4).assign(TRANM=0.1)))


@pytest.mark.skipif(not HAVE_OPM, reason="Requires OPM")
def test_main(tmp_path, mocker):
    """Test command line interface"""
    tmpcsvfile = tmp_path / "nnc.csv"
    mocker.patch("sys.argv", ["ecl2csv", "nnc", "-v", REEK, "-o", str(tmpcsvfile)])
    ecl2csv.main()

    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert not disk_df.empty
    assert "I1" in disk_df
    assert "TRAN" in disk_df


@pytest.mark.skipif(not HAVE_OPM, reason="Requires OPM")
def test_magic_stdout():
    """Test that we can pipe the output into a dataframe"""
    result = subprocess.run(
        ["ecl2csv", "nnc", "-o", "-", REEK], check=True, stdout=subprocess.PIPE
    )
    df_stdout = pd.read_csv(io.StringIO(result.stdout.decode()))
    assert not df_stdout.empty



"""Test module for ecl2df.trans"""

from pathlib import Path

import pytest

try:
    import networkx

    HAVE_NETWORKX = True
except ImportError:
    HAVE_NETWORKX = False

import pandas as pd

from ecl2df import ecl2csv, trans
from ecl2df.eclfiles import EclFiles

TESTDIR = Path(__file__).absolute().parent
REEK = str(TESTDIR / "data/reek/eclipse/model/2_R001_REEK-0.DATA")
EIGHTCELLS = str(TESTDIR / "data/eightcells/EIGHTCELLS.DATA")


def test_trans():
    """Test that we can build a dataframe of transmissibilities"""
    eclfiles = EclFiles(REEK)
    trans_df = trans.df(eclfiles)
    assert "TRAN" in trans_df
    assert "DIR" in trans_df
    assert set(trans_df["DIR"].unique()) == set(["I", "J", "K"])
    assert trans_df["TRAN"].sum() > 0

    trans_full_length = len(trans_df)

    # Try including some vectors:
    trans_df = trans.df(eclfiles, vectors="FIPNUM")
    assert "FIPNUM" not in trans_df
    assert "FIPNUM1" in trans_df
    assert "EQLNUM2" not in trans_df

    trans_df = trans.df(eclfiles, vectors=["FIPNUM", "EQLNUM"])
    assert "FIPNUM1" in trans_df
    assert "EQLNUM2" in trans_df

    trans_df = trans.df(eclfiles, vectors="BOGUS")
    assert "BOGUS1" not in trans_df
    assert "TRAN" in trans_df  # (we should have gotten a warning only)

    assert "K" not in trans.df(eclfiles, onlyijdir=True)["DIR"]
    assert "I" not in trans.df(eclfiles, onlykdir=True)["DIR"]

    # A warning is logged, seems strange to filter on both, but
    # the answer (empty) makes sense given the instruction. Alternative
    # would be a ValueError.
    assert trans.df(eclfiles, onlykdir=True, onlyijdir=True).empty

    transnnc_df = trans.df(eclfiles, addnnc=True)
    assert len(transnnc_df) > trans_full_length

    trans_df = trans.df(eclfiles, vectors=["FIPNUM", "EQLNUM"], boundaryfilter=True)
    assert trans_df.empty

    trans_df = trans.df(eclfiles, vectors="FIPNUM", boundaryfilter=True)
    assert len(trans_df) < trans_full_length

    trans_df = trans.df(eclfiles, coords=True)
    assert "X" in trans_df
    assert "Y" in trans_df


def test_grouptrans():
    """Test grouping of transmissibilities"""
    eclfiles = EclFiles(REEK)
    trans_df = trans.df(eclfiles, vectors="FIPNUM", group=True, coords=True)
    assert "FIPNUMPAIR" in trans_df
    assert "FIPNUM1" in trans_df
    assert "FIPNUM2" in trans_df
    assert (trans_df["FIPNUM1"] < trans_df["FIPNUM2"]).all()
    assert len(trans_df) == 7
    assert "X" in trans_df  # (average X coord for that FIPNUM interface)

    # This gives a logged error:
    assert trans.df(eclfiles, vectors=["FIPNUM", "EQLNUM"], group=True).empty


@pytest.mark.skipif(not HAVE_NETWORKX, reason="Requires networkx being installed")
def test_nx(tmp_path):
    """Test graph generation"""
    eclfiles = EclFiles(REEK)
    network = trans.make_nx_graph(eclfiles, region="FIPNUM")
    assert network.number_of_nodes() == 6
    networkx.write_gexf(network, tmp_path / "reek-fipnum-trans.gxf", prettyprint=True)
    assert (tmp_path / "reek-fipnum-trans.gxf").is_file()


def test_main(tmp_path, mocker):
    """Test command line interface"""
    tmpcsvfile = tmp_path / "trans.csv"
    mocker.patch(
        "sys.argv", ["ecl2csv", "trans", "-v", EIGHTCELLS, "-o", str(tmpcsvfile)]
    )
    ecl2csv.main()
    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert not disk_df.empty



"""Extract WCON* from an Eclipse deck"""

import argparse
import datetime
import logging
from typing import Union

import pandas as pd

try:
    # Needed for mypy

    # pylint: disable=unused-import
    import opm.io
except ImportError:
    pass

from ecl2df import EclFiles, getLogger_ecl2csv
from ecl2df.common import (
    parse_opmio_date_rec,
    parse_opmio_deckrecord,
    write_dframe_stdout_file,
)

logger = logging.getLogger(__name__)

# The keywords supported in this module.
WCONKEYS = ["WCONHIST", "WCONINJE", "WCONINJH", "WCONPROD"]


def df(deck: Union[EclFiles, "opm.libopmcommon_python.Deck"]) -> pd.DataFrame:
    """Loop through the deck and pick up information found

    The loop over the deck is a state machine, as it has to pick up dates
    """

    if isinstance(deck, EclFiles):
        deck = deck.get_ecldeck()

    wconrecords = []  # List of dicts of every line in input file
    date = None  # DATE columns will always be there, but can contain NaN
    for kword in deck:
        if kword.name in ["DATES", "START"]:
            for rec in kword:
                logger.info("Parsing at date %s", str(date))
                date = parse_opmio_date_rec(rec)
        elif kword.name == "TSTEP":
            if not date:
                logger.critical("Can't use TSTEP when there is no start_date")
                return pd.DataFrame()
            for rec in kword:
                steplist = rec[0].get_raw_data_list()
                # Assuming not LAB units, then the unit is days.
                days = sum(steplist)
                date += datetime.timedelta(days=days)
                logger.info(
                    "Advancing %s days to %s through TSTEP", str(days), str(date)
                )
        elif kword.name in WCONKEYS:
            for rec in kword:  # Loop over the lines inside WCON* record
                rec_data = parse_opmio_deckrecord(rec, kword.name)
                rec_data["DATE"] = date
                rec_data["KEYWORD"] = kword.name
                wconrecords.append(rec_data)

        elif kword.name == "TSTEP":
            logger.warning("WARNING: Possible premature stop at first TSTEP")
            break

    wcon_df = pd.DataFrame(wconrecords)

    return wcon_df


def fill_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Set up sys.argv parsers.

    Arguments:
        parser (argparse.ArgumentParser or argparse.subparser): parser to
            fill with arguments
    """
    parser.add_argument(
        "DATAFILE", help="Name of Eclipse DATA file or Eclipse include file."
    )
    parser.add_argument(
        "-o", "--output", type=str, help="Name of output csv file.", default="wcon.csv"
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Be verbose")
    return parser


def wcon_main(args) -> None:
    """Read from disk and write CSV back to disk"""
    logger = getLogger_ecl2csv(  # pylint: disable:redefined-outer_name
        __name__, vars(args)
    )
    eclfiles = EclFiles(args.DATAFILE)
    if eclfiles:
        deck = eclfiles.get_ecldeck()
    wcon_df = df(deck)
    write_dframe_stdout_file(
        wcon_df,
        args.output,
        index=False,
        caller_logger=logger,
        logstr=f"Wrote to {args.output}",
    )



"""Test module for pvt"""

import io
import os
import subprocess
from pathlib import Path

import pandas as pd
import pytest

from ecl2df import csv2ecl, ecl2csv, pvt
from ecl2df.eclfiles import EclFiles

try:
    import opm  # noqa
except ImportError:
    pytest.skip(
        "OPM is not installed, nothing relevant in here then",
        allow_module_level=True,
    )

TESTDIR = Path(__file__).absolute().parent
REEK = str(TESTDIR / "data/reek/eclipse/model/2_R001_REEK-0.DATA")
EIGHTCELLS = str(TESTDIR / "data/eightcells/EIGHTCELLS.DATA")


def test_pvto_strings():
    """Test PVTO parsing from strings"""
    pvto_deck = """PVTO
    0      1 1.0001 1
         200 1.000  1.001 /
    18    25 1.14  0.59 /
    /  -- One table (pvtnum=1), two records (two gor's)
    """
    dframe = pvt.pvto_fromdeck(EclFiles.str2deck(pvto_deck))
    assert "PVTNUM" in dframe
    assert "RS" in dframe
    assert "PRESSURE" in dframe
    assert "VISCOSITY" in dframe
    assert "VOLUMEFACTOR" in dframe
    assert len(dframe) == 3
    assert len(dframe["RS"].unique()) == 2
    assert len(dframe["PRESSURE"].unique()) == 3
    assert len(dframe["VOLUMEFACTOR"].unique()) == 3
    assert len(dframe["VISCOSITY"].unique()) == 3
    assert set(dframe["PVTNUM"].values) == {1}
    assert max(dframe["PRESSURE"]) == 200

    dframe_via_string = pvt.pvto_fromdeck(pvt.df2ecl_pvto(dframe))
    pd.testing.assert_frame_equal(dframe_via_string, dframe)

    # Provide TABDIMS in first test.. Infer later
    pvto_deck = """TABDIMS
     1 2 /

    PVTO
    0      1 1.0001 1
         200 1.000  1.001 /
    18    25 1.14  0.59 /
    /
    1      2 1.0001 1
         333 1.000  1.001 /
    19    30 1.14  0.59 /
    /
    """
    dframe = pvt.pvto_fromdeck(EclFiles.str2deck(pvto_deck))
    assert len(dframe) == 6
    assert "PVTNUM" in dframe
    assert set(dframe["PVTNUM"].astype(int).unique()) == {1, 2}
    assert len(dframe["RS"].unique()) == 4
    assert len(dframe["PRESSURE"].unique()) == 6
    assert len(dframe["VOLUMEFACTOR"].unique()) == 3

    dframe_via_string = pvt.pvto_fromdeck(pvt.df2ecl_pvto(dframe))
    pd.testing.assert_frame_equal(dframe_via_string, dframe)

    # Now test the same but without TABDIMS:
    pvto_deck = """
    PVTO
    0      1 1.0001 1
         200 1.000  1.001 /
    18    25 1.14  0.59 /
    /
    1      2 1.0001 1
         333 1.000  1.001 /
    19    30 1.14  0.59 /
    /
    """
    dframe = pvt.pvto_fromdeck(pvto_deck)
    assert len(dframe) == 6
    assert "PVTNUM" in dframe
    assert set(dframe["PVTNUM"].astype(int).unique()) == {1, 2}
    assert len(dframe["RS"].unique()) == 4
    assert len(dframe["PRESSURE"].unique()) == 6
    assert len(dframe["VOLUMEFACTOR"].unique()) == 3
    dframe_via_string = pvt.pvto_fromdeck(pvt.df2ecl_pvto(dframe))
    pd.testing.assert_frame_equal(dframe_via_string, dframe)

    # Test emtpy data:
    inc = pvt.df2ecl_pvto(pvt.df(""))
    assert "No data" in inc
    assert pvt.df(inc).empty


def test_pvdg_string():
    """Test that PVDG can be parsed from a string"""
    string = """
PVDG
400 6 0.01
600 3 0.012
1000 1.3 0.15 /
200 8 0.013
300 4 0.014
8000 1.8 0.16 /
"""
    dframe = pvt.pvdg_fromdeck(string)
    assert len(dframe) == 6
    assert "PVTNUM" in dframe
    assert len(dframe["PVTNUM"].unique()) == 2
    assert "PRESSURE" in dframe
    assert "VOLUMEFACTOR" in dframe
    assert "VISCOSITY" in dframe

    # Test emtpy data:
    inc = pvt.df2ecl_pvdg(pvt.df(""))
    assert "No data" in inc
    assert pvt.df(inc).empty


def test_pvdo_string():
    """Test that PVDO can be parsed from a string"""
    string = """
PVDO
400 6 0.01
1000 1.3 0.15 /
200 8 0.013
8000 1.8 0.16 /
"""
    dframe = pvt.pvdo_fromdeck(string)
    pd.testing.assert_frame_equal(
        dframe,
        pd.DataFrame(
            columns=["PRESSURE", "VOLUMEFACTOR", "VISCOSITY", "PVTNUM"],
            data=[
                [400.0, 6, 0.01, 1],
                [1000.0, 1.3, 0.15, 1],
                [200.0, 8, 0.013, 2],
                [8000.0, 1.8, 0.16, 2],
            ],
        ),
    )

    # Test emtpy data:
    inc = pvt.df2ecl_pvdo(pvt.df(""))
    assert "No data" in inc
    assert pvt.df(inc).empty


def test_pvt_reek():
    """Test that the Reek PVT input can be parsed individually"""

    eclfiles = EclFiles(REEK)
    pvto_df = pvt.pvto_fromdeck(eclfiles.get_ecldeck())
    assert "PVTNUM" in pvto_df
    assert "PRESSURE" in pvto_df
    assert "VOLUMEFACTOR" in pvto_df
    assert "VISCOSITY" in pvto_df
    assert max(pvto_df["PVTNUM"]) == 1
    assert max(pvto_df["PRESSURE"]) == 700.1
    # Check count of undersaturated lines pr. RS:
    # (nb: double brackets in .loc to ensure dataframe is returned)
    assert len(pvto_df.set_index("RS").loc[[0]]) == 2
    assert len(pvto_df.set_index("RS").loc[[15.906]]) == 1
    assert len(pvto_df.set_index("RS").loc[[105.5]]) == 15
    assert len(pvto_df["RS"].unique()) == 20
    assert pvto_df["VOLUMEFACTOR"].max() == 2.851
    assert pvto_df["VISCOSITY"].max() == 1.0001

    dframe_via_string = pvt.pvto_fromdeck(pvt.df2ecl_pvto(pvto_df))
    pd.testing.assert_frame_equal(dframe_via_string, pvto_df)

    density_df = pvt.density_fromdeck(eclfiles.get_ecldeck())
    pd.testing.assert_frame_equal(
        density_df,
        pd.DataFrame(
            columns=["PVTNUM", "OILDENSITY", "WATERDENSITY", "GASDENSITY"],
            data=[[1, 827.64, 999.04, 1.1427]],
        ),
        check_like=True,
    )
    dframe_via_string = pvt.density_fromdeck(pvt.df2ecl_density(density_df))
    pd.testing.assert_frame_equal(dframe_via_string, density_df)

    rock_df = pvt.rock_fromdeck(eclfiles.get_ecldeck())
    assert "PVTNUM" in rock_df
    assert len(rock_df) == 1
    assert "PRESSURE" in rock_df
    assert "COMPRESSIBILITY" in rock_df
    assert rock_df["PRESSURE"].values[0] == 327.3

    pvtw_df = pvt.pvtw_fromdeck(eclfiles.get_ecldeck())
    assert "PVTNUM" in pvtw_df
    assert pvtw_df["PVTNUM"].values[0] == 1
    assert len(pvtw_df) == 1
    assert "PRESSURE" in pvtw_df
    assert "VOLUMEFACTOR" in pvtw_df
    assert "COMPRESSIBILITY" in pvtw_df
    assert "VISCOSITY" in pvtw_df
    assert "VISCOSIBILITY" in pvtw_df
    assert pvtw_df["VISCOSITY"].values[0] == 0.25

    pvdg_df = pvt.pvdg_fromdeck(eclfiles.get_ecldeck())
    assert "PVTNUM" in pvdg_df
    assert "PRESSURE" in pvdg_df
    assert "VOLUMEFACTOR" in pvdg_df
    assert "VISCOSITY" in pvdg_df
    assert len(pvdg_df["PVTNUM"].unique()) == 1
    assert pvdg_df["PVTNUM"].max() == 1
    assert len(pvdg_df) == 15


def test_pvtg_string():
    """Test parsing of PVTG"""

    # Example data from E100 manual
    string = """
PVTG
30 0.00014    0.0523 0.0234
   0          0.0521 0.0238 /
90 0.00012    0.0132 0.0252
   0          0.0131 0.0253 /
150 0.00015   0.00877 0.0281
   0          0.00861 0.0275 /
210 0.00019   0.00554 0.0318
    0         0.00555 0.0302 /
270 0.00029   0.00417 0.0355
    0         0.00421 0.0330 /
330 0.00049   0.00357 0.0392
    0         0.00361 0.0358 /
530 0.00060   0.00356 0.0393
    0         0.00360 0.0359 /
/ null record to terminate table 1
60 0.00014    0.0523 0.0234 /
120 0.00012   0.0132 0.0252 /
180 0.00015   0.00877 0.0281 /
240 0.00019   0.00554 0.0318 /
300 0.00029   0.00417 0.0355 /
360 0.00049   0.00357 0.0392 /
560 0.00060   0.00356 0.0393
    0         0.00360 0.0359 / undersaturated data for Pg=560
/ null record to terminate table 2
"""
    pvtg_df = pvt.pvtg_fromdeck(string)
    assert "PRESSURE" in pvtg_df
    assert "OGR" in pvtg_df
    assert "PVTNUM" in pvtg_df
    assert "VOLUMEFACTOR" in pvtg_df
    assert "VISCOSITY" in pvtg_df
    assert len(pvtg_df["PVTNUM"].unique()) == 2
    assert len(pvtg_df["PRESSURE"].unique()) == 14
    assert max(pvtg_df["VOLUMEFACTOR"]) == 0.0523
    assert max(pvtg_df["VISCOSITY"]) == 0.0393

    # Test empty data:
    inc = pvt.df2ecl_pvtg(pvt.df(""))
    assert "No data" in inc
    assert pvt.df(inc).empty

    # Simpler string and dataframe
    string = """
PVTG
30 0.00014    0.0523 0.0234
    0         0.00360 0.0359 /
/
60 0.00014    0.0523 0.0234 /
/
"""
    df = pvt.pvtg_fromdeck(string)
    pd.testing.assert_frame_equal(
        df,
        pd.DataFrame(
            columns=["OGR", "VOLUMEFACTOR", "VISCOSITY", "PRESSURE", "PVTNUM"],
            data=[
                [0.00014, 0.0523, 0.0234, 30.0, 1],
                [0.00000, 0.0036, 0.0359, 30.0, 1],
                [0.00014, 0.0523, 0.0234, 60.0, 2],
            ],
        ),
    )


def test_density():
    """Test that DENSITY can be parsed from files and from strings"""
    eclfiles = EclFiles(REEK)
    density_df = pvt.density_fromdeck(eclfiles.get_ecldeck())
    assert len(density_df) == 1
    assert "PVTNUM" in density_df
    assert "OILDENSITY" in density_df
    assert "WATERDENSITY" in density_df
    assert "GASDENSITY" in density_df

    dframe_via_string = pvt.density_fromdeck(pvt.df2ecl_density(density_df))
    pd.testing.assert_frame_equal(dframe_via_string, density_df)

    two_pvtnum_deck = """DENSITY
        860      999.04       1.1427 /
        800      950     1.05
        /
        """
    density_df = pvt.density_fromdeck(EclFiles.str2deck(two_pvtnum_deck))
    # (a warning will be printed that we cannot guess)
    assert len(density_df) == 1
    density_df = pvt.density_fromdeck(two_pvtnum_deck)
    assert "PVTNUM" in density_df
    assert density_df["PVTNUM"].max() == 2
    assert density_df["PVTNUM"].min() == 1
    assert "OILDENSITY" in density_df
    dframe_via_string = pvt.density_fromdeck(pvt.df2ecl_density(density_df))
    pd.testing.assert_frame_equal(dframe_via_string, density_df)

    # Test emtpy data:
    inc = pvt.df2ecl_density(pvt.df(""))
    assert "No data" in inc
    assert pvt.df(inc).empty


def test_pvtw():
    """Test that PVTW can be parsed from a string"""
    deck = """PVTW
     327.3         1.03    4.51E-005         0.25            0 /"""
    pvtw_df = pvt.pvtw_fromdeck(EclFiles.str2deck(deck))
    pd.testing.assert_frame_equal(
        pvtw_df,
        pd.DataFrame(
            columns=[
                "PRESSURE",
                "VOLUMEFACTOR",
                "COMPRESSIBILITY",
                "VISCOSITY",
                "VISCOSIBILITY",
                "PVTNUM",
            ],
            data=[[327.3, 1.03, 4.51e-005, 0.25, 0.0, 1]],
        ),
        check_like=True,
    )

    deck = """PVTW
     327.3         1.03    4.51E-005         0.25            0 /
     300           1    0.0001  0.2 /"""
    pvtw_df = pvt.pvtw_fromdeck(deck)  # Must give string, not deck, for NTPVT guessing
    assert len(pvtw_df) == 2

    # Test emtpy data:
    inc = pvt.df2ecl_pvtw(pvt.df(""))
    assert "No data" in inc
    assert pvt.df(inc).empty


def test_rock():
    """Test parsing of the ROCK keyword from a string"""
    deck = """ROCK
     100 1.1 /"""
    rock_df = pvt.rock_fromdeck(EclFiles.str2deck(deck))
    assert len(rock_df) == 1
    assert "PRESSURE" in rock_df
    assert "COMPRESSIBILITY" in rock_df
    dframe_via_string = pvt.rock_fromdeck(pvt.df2ecl_rock(rock_df))
    pd.testing.assert_frame_equal(dframe_via_string, rock_df)

    # Test emtpy data:
    inc = pvt.df2ecl_rock(pvt.df(""))
    assert "No data" in inc
    assert pvt.df(inc).empty


def test_df():
    """Test that aggregate dataframes are produced"""
    eclfiles = EclFiles(REEK)
    pvtdf = pvt.df(eclfiles)

    assert not pvtdf.empty
    assert set(pvtdf["KEYWORD"]) == {"PVTO", "PVDG", "DENSITY", "ROCK", "PVTW"}
    assert "PVTNUM" in pvtdf
    assert "PRESSURE" in pvtdf
    assert "RS" in pvtdf
    assert "COMPRESSIBILITY" in pvtdf
    assert "VISCOSITY" in pvtdf
    assert len(pvtdf["PVTNUM"].unique()) == 1


def test_main(tmp_path, mocker):
    """Test command line interface"""
    os.chdir(tmp_path)
    tmpcsvfile = tmp_path / "pvt.csv"
    mocker.patch(
        "sys.argv", ["ecl2csv", "pvt", "-v", EIGHTCELLS, "-o", str(tmpcsvfile)]
    )
    ecl2csv.main()

    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(tmpcsvfile)
    assert "PVTNUM" in disk_df
    assert "KEYWORD" in disk_df
    assert not disk_df.empty

    # Write back to include file:
    incfile = tmp_path / "pvt.inc"
    mocker.patch(
        "sys.argv", ["csv2ecl", "pvt", "-v", str(tmpcsvfile), "-o", str(incfile)]
    )
    csv2ecl.main()

    # Reparse the include file on disk back to dataframe
    # and check dataframe equality
    assert Path(incfile).is_file()
    disk_inc_df = pvt.df(open(incfile).read())
    pd.testing.assert_frame_equal(disk_df, disk_inc_df)

    # Test entry point towards include file:
    (Path(tmp_path) / "pvto.inc").write_text(
        """PVTO
    0      1 1.0001 1
         200 1.000  1.001 /
    18    25 1.14  0.59 /
    /
    """
    )
    mocker.patch("sys.argv", ["ecl2csv", "pvt", "-v", "pvto.inc", "-o", "pvto.csv"])
    ecl2csv.main()
    assert Path("pvto.csv").is_file()

    # Empty data:
    (Path(tmp_path) / "empty.inc").write_text(
        """SWOF
    0      1 1.0001 1 /
    /
    """
    )
    mocker.patch("sys.argv", ["ecl2csv", "pvt", "-v", "empty.inc", "-o", "empty.csv"])
    ecl2csv.main()
    assert not Path("empty.csv").read_text().strip()


def test_magic_stdout(tmp_path):
    """Test writing dataframes and include files to stdout"""
    os.chdir(tmp_path)
    result = subprocess.run(
        ["ecl2csv", "pvt", "-o", "-", EIGHTCELLS], check=True, stdout=subprocess.PIPE
    )
    df_stdout = pd.read_csv(io.StringIO(result.stdout.decode()))
    assert not df_stdout.empty

    # Verbose options should not ruin it:
    result = subprocess.run(
        ["ecl2csv", "pvt", "--verbose", "-o", "-", EIGHTCELLS],
        check=True,
        stdout=subprocess.PIPE,
    )
    df_stdout = pd.read_csv(io.StringIO(result.stdout.decode()))
    assert not df_stdout.empty

    # Pipe back to csv2ecl:
    df_stdout.to_csv("pvt.csv", index=False)
    result = subprocess.run(
        ["csv2ecl", "pvt", "--verbose", "-o", "-", "pvt.csv"],
        check=True,
        stdout=subprocess.PIPE,
    )
    dframe = pvt.df(result.stdout.decode())
    assert not dframe.empty


def test_df2ecl():
    """df2ecl is a wrapper around the df2ecl_* functions

    The validity of produced dataframes is tested in other test functions
    herein, here we mainly test for the API and error handling"""
    with pytest.raises(ValueError):
        pvt.df2ecl(pd.DataFrame())


def test_df2ecl_pvto():
    """Test that we can print a PVTO dataframe to E100 include file"""
    dframe = pd.DataFrame(
        columns=["PVTNUM", "RS", "PRESSURE", "VOLUMEFACTOR", "VISCOSITY"],
        data=[[1, 50, 100, 2, 1.04]],
    )
    pvto_string = pvt.df2ecl_pvto(dframe)
    assert "PVTO" in pvto_string
    assert "1.04" in pvto_string
    assert "100" in pvto_string
    dframe_from_str = pvt.df(pvto_string)
    pd.testing.assert_frame_equal(
        dframe,
        dframe_from_str.drop("KEYWORD", axis="columns"),
        check_like=True,
        check_dtype=False,
    )

    dframe = pd.DataFrame(
        columns=["PVTNUM", "RS", "PRESSURE", "VOLUMEFACTOR", "VISCOSITY"],
        data=[[1, 50, 100, 2, 1.04], [1, 50, 120, 3, 1.05]],
    )
    pvto_string = pvt.df2ecl_pvto(dframe)
    assert "PVTO" in pvto_string
    assert "1.05" in pvto_string
    assert "120" in pvto_string
    dframe_from_str = pvt.df(pvto_string)
    pd.testing.assert_frame_equal(
        dframe,
        dframe_from_str.drop("KEYWORD", axis="columns"),
        check_like=True,
        check_dtype=False,
    )

    # If PVTNUM is missing, the code gives up if there are many rows.
    assert "PVTO" not in pvt.df2ecl_pvto(
        pd.concat([dframe, dframe]).drop("PVTNUM", axis="columns")
    )

    # If only one row, this is accepted:
    assert "PVTO" in pvt.df2ecl_pvto(dframe.head(1).drop("PVTNUM", axis="columns"))
    # (the corner case with only one row is not very meaningful, but at
    # least it is well defined how to treat it)


def test_df2ecl_rock(tmp_path):
    """Test generation of ROCK include files from dataframes"""
    os.chdir(tmp_path)

    rock_df = pd.DataFrame(
        columns=["PVTNUM", "KEYWORD", "PRESSURE", "COMPRESSIBILITY"],
        data=[[1, "ROCK", 100, 0.001]],
    )

    rock_inc = pvt.df2ecl(rock_df)
    assert "ROCK" in rock_inc
    rock_inc = pvt.df2ecl(rock_df, comments=dict(ROCK="foo"))
    assert "foo" in rock_inc
    rock_inc = pvt.df2ecl(rock_df, comments=dict(DENSITY="foo"))
    assert "foo" not in rock_inc

    rock_inc = pvt.df2ecl(rock_df, comments=dict(ROCK="foo\nbar"), filename="foo.inc")
    assert Path("foo.inc").is_file()
    assert "foo" in rock_inc
    assert "bar" in rock_inc
    # Multiline comments are tricky, is the output valid?
    rock_df_from_inc = pvt.rock_fromdeck(rock_inc).assign(KEYWORD="ROCK")
    # Need to sort columns for comparison, as column order does not matter
    # in dataframes, but it does in the function assert_frame_equal
    rock_df_from_inc = rock_df_from_inc.reindex(
        sorted(rock_df_from_inc.columns), axis=1
    )
    rock_df = rock_df_from_inc.reindex(sorted(rock_df.columns), axis=1)
    pd.testing.assert_frame_equal(rock_df_from_inc, rock_df)

    rock_inc = pvt.df2ecl(rock_df, keywords=["DENSITY"])
    assert not rock_inc
    rock_inc = pvt.df2ecl(rock_df, keywords="DENSITY")
    assert not rock_inc

    rock_inc = pvt.df2ecl(rock_df, keywords=["ROCK", "DENSITY"])
    assert "ROCK" in rock_inc
    assert "DENSITY" not in rock_inc
    rock_inc = pvt.df2ecl(rock_df, keywords="ROCK")
    assert "ROCK" in rock_inc

    # This dataframe is ignored, if we miss PVTNUM:
    ambig_rock_df = pd.DataFrame(
        columns=["KEYWORD", "PRESSURE", "COMPRESSIBILITY"],
        data=[["ROCK", 100, 0.001], ["ROCK", 200, 0.002]],
    )
    assert "ROCK" not in pvt.df2ecl_rock(ambig_rock_df)

    # But if only one row, it is ok:
    assert "ROCK" in pvt.df2ecl_rock(ambig_rock_df.head(1))

    # If we don't want the ROCK keyword, we won't get it:
    nonrock_inc = pvt.df2ecl(rock_df, keywords=["PVTO"])
    assert "ROCK" not in nonrock_inc


def test_df2ecl_density():
    """Test generation of PVT density include files from dataframes"""
    density_df = pd.DataFrame(
        columns=["PVTNUM", "OILDENSITY", "WATERDENSITY", "GASDENSITY"],
        data=[[1, 827.64, 999.04, 1.1427]],
    )

    dens_inc = pvt.df2ecl_density(density_df)
    assert "DENSITY" in dens_inc

    # If PVTNUM is missing, the code gives up:
    assert "DENSITY" not in pvt.df2ecl_density(
        pd.concat([density_df, density_df]).drop("PVTNUM", axis="columns")
    )

    # Unless there is only one row:
    assert "DENSITY" in pvt.df2ecl_density(density_df.drop("PVTNUM", axis="columns"))

    # Missing column:
    with pytest.raises(KeyError, match="OILDENSITY"):
        pvt.df2ecl_density(density_df.drop("OILDENSITY", axis="columns"))


def test_df2ecl_pvtw():
    """Test generation of PVTW include statements"""
    pvtw_df = pd.DataFrame(
        columns=[
            "PRESSURE",
            "VOLUMEFACTOR",
            "COMPRESSIBILITY",
            "VISCOSITY",
            "VISCOSIBILITY",
            "PVTNUM",
        ],
        data=[[327.3, 1.03, 4.51e-005, 0.25, 0.0, 1]],
    )
    assert "PVTW" in pvt.df2ecl_pvtw(pvtw_df)

    # If PVTNUM is missing, the code gives up:
    assert "PVTW" not in pvt.df2ecl_pvtw(
        pd.concat([pvtw_df, pvtw_df]).drop("PVTNUM", axis="columns")
    )

    # Unless there is only one row:
    assert "PVTW" in pvt.df2ecl_pvtw(pvtw_df.drop("PVTNUM", axis="columns"))

    # Missing column:
    with pytest.raises(KeyError, match="VOLUMEFACTOR"):
        pvt.df2ecl_pvtw(pvtw_df.drop("VOLUMEFACTOR", axis="columns"))


def test_df2ecl_pvtg():
    """Test generation of PVTG include statements"""
    pvtg_df = pd.DataFrame(
        columns=["OGR", "VOLUMEFACTOR", "VISCOSITY", "PRESSURE", "PVTNUM"],
        data=[
            [0.00014, 0.0523, 0.0234, 30.0, 1],
            [0.00000, 0.0036, 0.0359, 30.0, 1],
            [0.00014, 0.0523, 0.0234, 60.0, 2],
        ],
    )
    assert "PVTG" in pvt.df2ecl_pvtg(pvtg_df)
    assert "PVTG" in pvt.df2ecl_pvtg(pvtg_df.assign(KEYWORD="PVTG"))
    pd.testing.assert_frame_equal(
        pvt.df(pvt.df2ecl_pvtg(pvtg_df)).drop("KEYWORD", axis="columns"), pvtg_df
    )

    # If PVTNUM is missing, the code gives up:
    assert "PVTG" not in pvt.df2ecl_pvtg(
        pd.concat([pvtg_df, pvtg_df]).drop("PVTNUM", axis="columns")
    )

    # Unless there is only one row:
    assert "PVTG" in pvt.df2ecl_pvtg(pvtg_df.head(1).drop("PVTNUM", axis="columns"))

    # Missing column:
    with pytest.raises(KeyError, match="VOLUMEFACTOR"):
        pvt.df2ecl_pvtg(pvtg_df.drop("VOLUMEFACTOR", axis="columns"))


def test_df2ecl_pvdo_pvdg():
    """Test construction of PVDO and PVDG statements from dataframe.

    The keyword data and code is similar enough to warrant one test
    for both functions, with the same dataset."""
    pvdog_df = pd.DataFrame(
        columns=["PRESSURE", "VOLUMEFACTOR", "VISCOSITY", "PVTNUM"],
        data=[
            [400.0, 6, 0.01, 1],
            [1000.0, 1.3, 0.15, 1],
            [200.0, 8, 0.013, 2],
            [8000.0, 1.8, 0.16, 2],
        ],
    )

    assert "PVDO" in pvt.df2ecl_pvdo(pvdog_df)
    assert "PVDG" in pvt.df2ecl_pvdg(pvdog_df)

    assert "PVDO" in pvt.df2ecl_pvdo(pvdog_df.assign(KEYWORD="PVDO"))
    assert "PVDG" in pvt.df2ecl_pvdg(pvdog_df.assign(KEYWORD="PVDG"))

    pd.testing.assert_frame_equal(
        pvt.df(pvt.df2ecl_pvdo(pvdog_df)).drop("KEYWORD", axis="columns"), pvdog_df
    )
    pd.testing.assert_frame_equal(
        pvt.df(pvt.df2ecl_pvdg(pvdog_df)).drop("KEYWORD", axis="columns"), pvdog_df
    )

    # If PVTNUM is missing, the code gives up:
    assert "PVDO" not in pvt.df2ecl_pvdo(
        pd.concat([pvdog_df, pvdog_df]).drop("PVTNUM", axis="columns")
    )
    assert "PVDG" not in pvt.df2ecl_pvdg(
        pd.concat([pvdog_df, pvdog_df]).drop("PVTNUM", axis="columns")
    )

    # Unless there is only one row:
    assert "PVDO" in pvt.df2ecl_pvdo(pvdog_df.head(1).drop("PVTNUM", axis="columns"))
    assert "PVDG" in pvt.df2ecl_pvdg(pvdog_df.head(1).drop("PVTNUM", axis="columns"))

    # Missing column:
    with pytest.raises(KeyError, match="VOLUMEFACTOR"):
        pvt.df2ecl_pvdo(pvdog_df.drop("VOLUMEFACTOR", axis="columns"))
    with pytest.raises(KeyError, match="VOLUMEFACTOR"):
        pvt.df2ecl_pvdg(pvdog_df.drop("VOLUMEFACTOR", axis="columns"))



"""Check that ecl2df's submodules are always imported"""
import sys

# This file tests what happens when we do this import:
import ecl2df


def test_init():
    assert "ecl2df.compdat" in sys.modules

    # This should be a list of all submodules
    assert ecl2df.SUBMODULES

    for submodule in ecl2df.SUBMODULES:
        assert "ecl2df." + submodule in sys.modules

    # The Eclfiles object inside eclfiles should be lifted up to top-level:
    assert hasattr(ecl2df, "EclFiles")

    assert isinstance(ecl2df.__version__, str)



"""Test module for rft"""
import datetime
import os
import random
from pathlib import Path

import pandas as pd
import pytest

from ecl2df import ecl2csv, rft
from ecl2df.eclfiles import EclFiles

TESTDIR = Path(__file__).absolute().parent
REEK = str(TESTDIR / "data/reek/eclipse/model/2_R001_REEK-0.DATA")
EIGHTCELLS = str(TESTDIR / "data/eightcells/EIGHTCELLS.DATA")

# pylint: disable=protected-access


def test_rftrecords2df():
    """Test that we can construct a dataframe for navigating in RFT
    records"""
    rftrecs = rft._rftrecords2df(EclFiles(EIGHTCELLS).get_rftfile())
    assert len(rftrecs[rftrecs["recordname"] == "TIME"]) == len(
        rftrecs["timeindex"].unique()
    )
    assert set(rftrecs["recordtype"].unique()) == set(["REAL", "INTE", "CHAR"])
    assert rftrecs["timeindex"].dtype == int
    assert rftrecs["recordidx"].dtype == int

    # Test that we have a consecutive index in "recordidx"
    assert (rftrecs["recordidx"] == rftrecs.index).all()


def test_rftrecords_generator():
    """Test the generator that will iterate over an EclFile/RFTFile
    and provide one yield pr.  well pr. date"""
    for rftrecord in rft.rftrecords(EclFiles(EIGHTCELLS).get_rftfile()):
        assert isinstance(rftrecord, dict)
        assert "date" in rftrecord
        assert isinstance(rftrecord["date"], datetime.date)
        assert "wellmodel" in rftrecord
        assert "timeindex" in rftrecord
        assert "wellname" in rftrecord
        assert len(rftrecord["wellname"]) < 9  # E100
        assert rftrecord["wellmodel"] in ["STANDARD", "MULTISEG"]
        assert "headers" in rftrecord
        assert isinstance(rftrecord["headers"], pd.DataFrame)


def test_get_con_seg_data():
    """Get CON data. Later add more code here to defend the name"""
    rftfile = EclFiles(EIGHTCELLS).get_rftfile()

    # Test the first record, it is a CON type (not multisegment)
    rftrecord = next(rft.rftrecords(rftfile))
    con_data = rft.get_con_seg_data(rftrecord, rftfile, "CON")
    assert {"CONIPOS", "CONJPOS", "CONKPOS", "DEPTH", "PRESSURE", "SWAT"}.issubset(
        set(con_data.columns)
    )
    assert all(con_data["CONIDX"] == con_data.index + 1)

    with pytest.raises(ValueError):
        rft.get_con_seg_data(None, None, "FOO")


def test_minimal_well():
    """Test a dummy well dataset

    |    segidx 1

    """
    one_seg = pd.DataFrame(
        {"SEGIDX": [1], "SEGNXT": [None], "SEGBRNO": [1], "SEGPRES": [195.8]}
    )
    with pytest.raises(ValueError, match="Insufficient topology"):
        rft.process_seg_topology(one_seg.drop("SEGIDX", axis="columns"))

    m_one_seg = rft.process_seg_topology(one_seg)
    assert m_one_seg["LEAF"][0]
    assert len(m_one_seg) == 1
    assert rft.count_wellbranches(one_seg) == 1
    assert rft.split_seg_icd(one_seg)[1].empty

    con_data = pd.DataFrame({"CONSEGNO": [1], "PRESSURE": [200.1], "CONPRES": [196.0]})
    con_seg = rft.merge_icd_seg_conseg(con_data, m_one_seg)
    assert len(con_seg) == 1
    assert "CONSEGNO" in con_seg
    assert "SEGIDX" in con_seg
    con_seg = rft.add_extras(con_seg)
    assert "COMPLETION_DP" in con_seg
    assert con_seg["COMPLETION_DP"].values[0] == 196.0 - 195.8
    assert con_seg["DRAWDOWN"].values[0] == 200.1 - 196.0
    assert rft.seg2dicttree(m_one_seg) == {1: {}}
    print(rft.pretty_print_well(m_one_seg))


def test_minimal_branched_well():
    r"""
     |       segidx 1
    / \      segidx 2 and 3
    """
    two_branch = pd.DataFrame(
        {"SEGIDX": [1, 2, 3], "SEGNXT": [None, 1, 1], "SEGBRNO": [1, 1, 2]}
    )
    con_data = pd.DataFrame(
        {"CONSEGNO": [2, 3], "PRESSURE": [301, 302], "CONPRES": [291, 292]}
    )
    m_two_branch = rft.process_seg_topology(two_branch)
    assert len(m_two_branch) == 4  # One extra row for the junction segment
    assert sum(m_two_branch["LEAF"]) == 2
    assert rft.count_wellbranches(m_two_branch) == 2
    assert rft.split_seg_icd(two_branch)[1].empty
    con_seg = rft.merge_icd_seg_conseg(con_data, m_two_branch)

    # Junction segment has no reservoir connection and is not included
    # in the merge.
    assert len(con_seg) == 2

    assert rft.seg2dicttree(m_two_branch) == {1: {2: {}, 3: {}}}
    # Junction segment points to two upstream segments:
    assert set(
        m_two_branch[m_two_branch["SEGIDX"] == 1]["SEGIDX_upstream"].astype(int)
    ) == {2, 3}
    assert int(m_two_branch.loc[0, "SEGIDX_upstream"]) == 2
    assert int(m_two_branch.loc[1, "SEGIDX_upstream"]) == 3


def test_single_branch_icd():
    """
    Test that we are able to untangle segment dataframes with ICDs
    modelled as individual brances.
    """

    # Single branch well, one mother segment, two connections,
    # one icd to each connection:

    #     |   segidx 1
    #     | - *  segidx 2 and 4, and reservoir
    #     | - *  segidx 3 and 5, and reservoir

    # legend: | = tubing
    #         - = icd
    #         * = reservoir connection
    wellseg = pd.DataFrame(
        {
            "SEGIDX": [1, 2, 3, 4, 5],
            "SEGNXT": [None, 1, 2, 2, 3],
            "SEGBRNO": [1, 1, 1, 2, 3],
        }
    )
    con_data = pd.DataFrame(
        {"CONSEGNO": [4, 5], "PRESSURE": [301, 302], "CONPRES": [291, 292]}
    )
    (seg_data, icd_data) = rft.split_seg_icd(wellseg)
    print(rft.seg2dicttree(wellseg))
    print(rft.pretty_print_well(wellseg))
    assert rft.count_wellbranches(seg_data) == 1

    assert len(icd_data) == 2
    assert all(icd_data.columns.str.startswith("ICD"))
    assert all(icd_data["ICD_SEGIDX"].values == [4, 5])
    assert all(icd_data["ICD_SEGBRNO"].values == [2, 3])
    assert all(icd_data["ICD_SEGBRNO_upstream"].values == [0, 0])

    con_seg = rft.merge_icd_seg_conseg(con_data, seg_data, icd_data)
    assert len(con_seg) == 2
    con_seg = rft.add_extras(con_seg)
    assert all(con_seg["DRAWDOWN"].values == [10, 10])


def test_single_branch_partly_icd():
    r"""
    Test that we are able to untangle segment dataframes with ICDs
    modelled as individual branches, but not on all reservoir connections

    Single branch well, one mother segment, two connections,
    one icd to one connection:

         |   segidx 1
         | - *  segidx 2 and 4 (icd), and reservoir
         | *  segidx 3, and reservoir

    legend: | = tubing
            - = icd
            * = reservoir connection


    This is the same layout as
       |  segidx 1
       |  segidx 2
      / \  segidx 3 and 4
     *   *
    which is a two-branch well. It is not possible to
    separate these two based on topology. Current code interprets
    the topology as the latter case.
    The ambiguity arises due to the assumption that ICD segments
    are on their on branches with only one segment. In real modelled
    wells, this assumption will be valid, as the tubing segments will
    consist of more segments than 1 (except corner cases)
    (check the LONELYSEG segment toplogy metadata column which
    is used in determining whether a segment is ICD or not)

    """
    wellseg = pd.DataFrame(
        {"SEGIDX": [1, 2, 3, 4], "SEGNXT": [None, 1, 2, 2], "SEGBRNO": [1, 1, 1, 2]}
    )
    con_data = pd.DataFrame(
        {"CONSEGNO": [4, 3], "PRESSURE": [301, 302], "CONPRES": [291, 292]}
    )
    (seg_data, icd_data) = rft.split_seg_icd(wellseg)
    print(rft.seg2dicttree(wellseg))
    print(rft.pretty_print_well(wellseg))
    assert rft.count_wellbranches(seg_data) == 2

    assert len(icd_data) == 0

    con_seg = rft.merge_icd_seg_conseg(con_data, seg_data, icd_data)
    assert len(con_seg) == 2
    con_seg = rft.add_extras(con_seg)
    assert all(con_seg["DRAWDOWN"].values == [10, 10])


def test_branched_icd_well():
    r"""Simplest possible branched well with ICD segments

         |          segidx 1
    * - / \ - *     segidx 2 and 3
    """
    wellseg = pd.DataFrame(
        {
            "SEGIDX": [1, 2, 3, 4, 5],
            "SEGNXT": [None, 1, 1, 2, 3],
            "SEGBRNO": [1, 1, 2, 3, 4],
        }
    )
    con_data = pd.DataFrame(
        {"CONSEGNO": [4, 5], "PRESSURE": [301, 302], "CONPRES": [291, 292]}
    )
    (seg_data, icd_data) = rft.split_seg_icd(wellseg)
    print(rft.seg2dicttree(wellseg))
    print(rft.pretty_print_well(wellseg))

    assert len(icd_data) == 2
    assert all(icd_data.columns.str.startswith("ICD"))
    assert all(icd_data["ICD_SEGIDX"].values == [4, 5])
    assert all(icd_data["ICD_SEGBRNO"].values == [3, 4])
    assert all(icd_data["ICD_SEGBRNO_upstream"].values == [0, 0])

    print(seg_data)
    assert rft.count_wellbranches(seg_data) == 2

    con_seg = rft.merge_icd_seg_conseg(con_data, seg_data, icd_data)
    print(con_seg)
    assert len(con_seg) == 2
    con_seg = rft.add_extras(con_seg)
    assert all(con_seg["DRAWDOWN"].values == [10, 10])


def test_longer_branched_icd_well():
    r"""Test a well with two connections on each of two laterals,
    with an ICD segment for each connection

           |          segidx 1
      * - / \ - *     segidx 4 and 2 (lateral1) and 6 and 8
      * - | | - *     segidx 5 and 3 (lateral1) and 7 and 9

    """
    wellseg = {
        "SEGIDX": [1] + [2, 3, 4, 5] + [6, 7, 8, 9],
        "SEGNXT": [None] + [1, 2, 2, 3] + [1, 6, 6, 7],
        "SEGBRNO": [1] + [1, 1, 3, 4] + [2, 2, 5, 6],
    }

    # Shuffle the segment list randomly, that should not matter:
    shuffled = list(range(9))
    random.shuffle(shuffled)
    print(shuffled)
    wellseg = pd.DataFrame(
        {segname: [wellseg[segname][idx] for idx in shuffled] for segname in wellseg}
    )

    con_data = pd.DataFrame(
        {
            "CONSEGNO": [4, 5, 8, 9],
            "PRESSURE": [301, 302, 401, 402],
            "CONPRES": [291, 292, 392, 393],
        }
    )
    seg_data = rft.process_seg_topology(wellseg)
    assert sum(seg_data["LONELYSEG"]) == 4
    assert sum(seg_data["LEAF"]) == 4
    assert sum(seg_data["JUNCTION"]) == 6  # 1, 2 and 6 counted twice
    assert sum(seg_data["LEAF"]) == 4
    (seg_data, icd_data) = rft.split_seg_icd(wellseg)
    print(rft.seg2dicttree(wellseg))
    print(rft.pretty_print_well(wellseg))

    assert len(icd_data) == 4
    assert all(icd_data.columns.str.startswith("ICD"))
    assert set(icd_data["ICD_SEGIDX"].values) == {4, 5, 8, 9}
    assert set(icd_data["ICD_SEGBRNO"].values) == {3, 4, 5, 6}
    assert all(icd_data["ICD_SEGBRNO_upstream"].values == [0, 0, 0, 0])

    print(seg_data)
    assert rft.count_wellbranches(seg_data) == 2

    con_seg = rft.merge_icd_seg_conseg(con_data, seg_data, icd_data)
    print(con_seg)
    assert len(con_seg) == 4
    con_seg = rft.add_extras(con_seg)
    assert all(con_seg["DRAWDOWN"].values == [10, 10, 9, 9])


def test_longer_branched_partly_icd_well():
    r"""Test a well with two connections on each of two laterals,
    with an ICD segment for each connection only on the first lateral

           |          segidx 1
      * - / \  *     segidx 4 and 2 (lateral1) and 6 and
      * - | |  *     segidx 5 and 3 (lateral1) and 7 and

    """
    wellseg = pd.DataFrame(
        {
            # [wellhead] + [lateral 1, tubing, then icd layer] + [lateral 2]
            "SEGIDX": [1] + [2, 3, 4, 5] + [6, 7],
            "SEGNXT": [None] + [1, 2, 2, 3] + [1, 6],
            "SEGBRNO": [1] + [1, 1, 3, 4] + [2, 2],
        }
    )

    con_data = pd.DataFrame(
        {
            "CONSEGNO": [4, 5, 6, 7],
            "PRESSURE": [301, 302, 401, 402],
            "CONPRES": [291, 292, 392, 393],
        }
    )
    (seg_data, icd_data) = rft.split_seg_icd(wellseg)
    print()
    print(seg_data)
    print(rft.seg2dicttree(wellseg))
    print(rft.pretty_print_well(wellseg))

    assert len(icd_data) == 2
    assert all(icd_data.columns.str.startswith("ICD"))
    assert set(icd_data["ICD_SEGIDX"].values) == {4, 5}
    assert set(icd_data["ICD_SEGBRNO"].values) == {3, 4}
    assert all(icd_data["ICD_SEGBRNO_upstream"].values == [0, 0])

    print(seg_data)
    assert rft.count_wellbranches(seg_data) == 2

    con_seg = rft.merge_icd_seg_conseg(con_data, seg_data, icd_data)
    print(con_seg)
    assert len(con_seg) == 4
    con_seg = rft.add_extras(con_seg)
    assert all(con_seg["DRAWDOWN"].values == [10, 10, 9, 9])


def test_seg2dicttree():
    assert rft.seg2dicttree(pd.DataFrame()) == {}
    with pytest.raises(ValueError):
        rft.seg2dicttree(pd.DataFrame({"SEGIDX": [1]}))

    with pytest.raises(KeyError, match="SEGBRNO"):
        rft.seg2dicttree(pd.DataFrame({"SEGIDX": [1], "SEGNXT": [None]}))

    # Simplest well:
    assert rft.seg2dicttree(
        pd.DataFrame({"SEGIDX": [1], "SEGNXT": [None], "SEGBRNO": [1]})
    ) == {1: {}}

    # Two branches:
    assert rft.seg2dicttree(
        pd.DataFrame(
            {"SEGIDX": [1, 2, 3], "SEGNXT": [None, 1, 1], "SEGBRNO": [1, 1, 2]}
        )
    ) == {1: {2: {}, 3: {}}}


@pytest.mark.parametrize(
    "dframe, inplace,  expected",
    [
        (pd.DataFrame(), True, pd.DataFrame()),
        (pd.DataFrame(), False, pd.DataFrame()),
        (
            pd.DataFrame([{"CONPRES": 30, "SEGPRES": 20}]),
            True,
            pd.DataFrame(
                [{"CONPRES": 30, "SEGPRES": 20, "COMPLETION_DP": 10, "DRAWDOWN": 0}]
            ),
        ),
        (
            pd.DataFrame([{"CONPRES": 30, "SEGPRES": 20}]),
            False,
            pd.DataFrame(
                [{"CONPRES": 30, "SEGPRES": 20, "COMPLETION_DP": 10, "DRAWDOWN": 0}]
            ),
        ),
        (
            pd.DataFrame([{"CONPRES": 30, "PRESSURE": 40}]),
            True,
            pd.DataFrame(
                [{"CONPRES": 30, "PRESSURE": 40, "DRAWDOWN": 10, "CONBPRES": 40}]
            ),
        ),
        (
            # Compute connection length
            pd.DataFrame([{"CONLENEN": 4, "CONLENST": 3}]),
            True,
            pd.DataFrame(
                [
                    {
                        "CONLENEN": 4,
                        "CONLENST": 3,
                        "CONMD": 3.5,
                        "CONLENTH": 1,
                        "DRAWDOWN": 0,
                    }
                ]
            ),
        ),
        (
            # Compute scaled rates (pr meter connection)
            pd.DataFrame([{"CONORAT": 400, "CONLENTH": 2}]),
            True,
            pd.DataFrame(
                [{"CONORAT": 400, "CONLENTH": 2, "CONORATS": 200.0, "DRAWDOWN": 0}]
            ),
        ),
        (
            pd.DataFrame([{"CONWRAT": 400, "CONLENTH": 2}]),
            True,
            pd.DataFrame(
                [{"CONWRAT": 400, "CONLENTH": 2, "CONWRATS": 200.0, "DRAWDOWN": 0}]
            ),
        ),
        (
            pd.DataFrame([{"CONGRAT": 400, "CONLENTH": 2}]),
            True,
            pd.DataFrame(
                [{"CONGRAT": 400, "CONLENTH": 2, "CONGRATS": 200.0, "DRAWDOWN": 0}]
            ),
        ),
    ],
)
def test_add_extras(dframe, inplace, expected):
    """Test addition of nice-to-have extras column."""
    original = dframe.copy()
    result = rft.add_extras(dframe, inplace)
    print(result)
    pd.testing.assert_frame_equal(result, expected, check_like=True)
    if inplace:
        pd.testing.assert_frame_equal(result, dframe)
    else:
        pd.testing.assert_frame_equal(dframe, original)


def test_rft2df():
    """Test that dataframes are produced"""
    eclfiles = EclFiles(REEK)
    rftdf = rft.df(eclfiles)
    assert "ZONE" in rftdf
    assert "LEAF" not in rftdf  # Topology metadata should not be exported
    assert set(rftdf["WELLMODEL"]) == {"STANDARD"}
    assert set(rftdf["WELL"]) == {
        "OP_1",
        "OP_2",
        "OP_3",
        "OP_4",
        "OP_5",
        "WI_1",
        "WI_2",
        "WI_3",
    }
    assert not rftdf.empty
    assert len(rftdf) == 115
    # Each well has 14 or 15 reservoir connections (14 layers in grid)
    assert set(rftdf.groupby("WELL")["CONIDX"].count().values) == {14, 15}
    assert not rftdf.columns.empty


def test_main_subparsers(tmp_path, mocker):
    """Test command line interface"""
    tmpcsvfile = tmp_path / ".TMP-rft.csv"
    mocker.patch("sys.argv", ["ecl2csv", "rft", EIGHTCELLS, "-o", str(tmpcsvfile)])
    ecl2csv.main()

    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert not disk_df.empty

    tmpcsvfile = tmp_path / ".TMP-rft2.csv"
    # Test with RFT file as argument:
    mocker.patch(
        "sys.argv",
        [
            "ecl2cvsv",
            "rft",
            "-v",
            REEK.replace(".DATA", ".RFT"),
            "-o",
            str(tmpcsvfile),
        ],
    )
    ecl2csv.main()
    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert not disk_df.empty


def test_main_debugmode(tmp_path, mocker):
    """Test debug mode"""
    os.chdir(tmp_path)
    mocker.patch(
        "sys.argv", ["ecl2csv", "rft", "--debug", EIGHTCELLS, "-o", "indebugmode.csv"]
    )
    ecl2csv.main()
    # Extra files emitted in debug mode:
    assert not pd.read_csv("con.csv").empty
    assert Path("seg.csv").exists()  # too simple example data, no segments.
    assert Path("icd.csv").exists()  # too simple example data, no ICD



"""Test module for compdat"""

from pathlib import Path

import pandas as pd
import pytest

from ecl2df import EclFiles, compdat, ecl2csv

try:
    import opm  # noqa
except ImportError:
    pytest.skip(
        "OPM is not installed",
        allow_module_level=True,
    )


TESTDIR = Path(__file__).absolute().parent
REEK = str(TESTDIR / "data/reek/eclipse/model/2_R001_REEK-0.DATA")
EIGHTCELLS = str(TESTDIR / "data/eightcells/EIGHTCELLS.DATA")

SCHFILE = str(TESTDIR / "data/reek/eclipse/include/schedule/reek_history.sch")

# Reek cases with multisegment well OP_6 including
# AICD and ICD completion from WellBuilder
SCHFILE_AICD = str(TESTDIR / "data/reek/eclipse/include/schedule/op6_aicd1_gp.sch")
SCHFILE_ICD = str(TESTDIR / "data/reek/eclipse/include/schedule/op6_icd1_gp.sch")
SCHFILE_VALV = str(TESTDIR / "data/reek/eclipse/include/schedule/op6_valve1_gp.sch")


def test_df():
    """Test main dataframe API, only testing that something comes out"""
    eclfiles = EclFiles(EIGHTCELLS)
    compdat_df = compdat.df(eclfiles)
    assert not compdat_df.empty
    assert "ZONE" in compdat_df
    assert "K1" in compdat_df
    assert "WELL" in compdat_df


def test_comp2df():
    """Test that dataframes are produced"""
    eclfiles = EclFiles(EIGHTCELLS)
    compdfs = compdat.deck2dfs(eclfiles.get_ecldeck())

    assert not compdfs["COMPDAT"].empty
    assert not compdfs["WELSEGS"].empty
    assert not compdfs["COMPSEGS"].empty
    assert not compdfs["COMPDAT"].columns.empty


def test_schfile2df():
    """Test that we can process individual files"""
    deck = EclFiles.file2deck(SCHFILE)
    compdfs = compdat.deck2dfs(deck)
    assert not compdfs["COMPDAT"].columns.empty
    assert not compdfs["COMPDAT"].empty


def test_str_compdat():
    """Test compdat parsing directly on strings"""
    schstr = """
COMPDAT
 'OP1' 33 110 31 31 'OPEN' 1* 6467.31299 0.216 506642.25  0 1* 'Y' 7.18 /
-- comments.
/
"""
    deck = EclFiles.str2deck(schstr)
    compdfs = compdat.deck2dfs(deck)
    compdat_df = compdfs["COMPDAT"]
    assert compdat_df.loc[0, "SATN"] == 0
    assert not compdat_df.loc[0, "DFACT"]
    assert compdat_df.loc[0, "DIR"] == "Y"

    schstr = """
COMPDAT
 'FOO' 303 1010 031 39  /
/
"""
    compdat_df = compdat.deck2dfs(EclFiles.str2deck(schstr))["COMPDAT"]
    assert len(compdat_df) == 9
    assert not compdat_df["DFACT"].values[0]
    assert not compdat_df["TRAN"].values[0]
    assert compdat_df["I"].values[0] == 303


def test_str2df():
    """Testing making a dataframe from an explicit string"""
    schstr = """
WELSPECS
 'OP1' 'OPWEST' 41 125 1759.74 'OIL' 0.0 'STD' 'SHUT' 'YES'  0  'SEG' /
/

COMPDAT
 'OP1' 33 110 31 31 'OPEN' 0 6467.31299 0.216 506642.25  0.0 0.0 'Y' 7.18 /
-- comments.
/

WELSEGS
  'OP1' 1689 1923 1.0E-5 'ABS' 'HFA' 'HO' / comment without -- identifier
-- foo bar
   2 2 1 1 1923.9 1689.000 0.1172 0.000015  /
/

COMPSEGS
  'OP1' / -- Yet a comment
  -- comment
  41 125 29  5 2577.0 2616.298 / icd on branch 1 in segment 17
/
-- (WSEGVALS is not processed)
WSEGVALV
  'OP1'   166   1   7.4294683E-06  0 / icd on segment 17, cell 41 125 29
/
"""
    deck = EclFiles.str2deck(schstr)
    compdfs = compdat.deck2dfs(deck)
    compdat_df = compdfs["COMPDAT"]
    welsegs = compdfs["WELSEGS"]
    compsegs = compdfs["COMPSEGS"]
    assert "WELL" in compdat_df
    assert len(compdat_df) == 1
    assert compdat_df["WELL"].unique()[0] == "OP1"

    # Check that we have not used the very long opm.io term here:
    assert "CONNECTION_TRANSMISSIBILITY_FACTOR" not in compdat_df
    assert "TRAN" in compdat_df

    assert "Kh" not in compdat_df  # Mixed-case should not be used.
    assert "KH" in compdat_df

    # Make sure the ' are ignored:
    assert compdat_df["OP/SH"].unique()[0] == "OPEN"

    # Continue to WELSEGS
    assert len(welsegs) == 1  # First record is appended to every row.

    # Since we have 'ABS' in WELSEGS, there should be an extra
    # column called 'SEGMENT_MD'
    assert "SEGMENT_MD" in welsegs
    assert welsegs["SEGMENT_MD"].max() == 1923.9

    # Test COMPSEGS
    assert len(compsegs) == 1
    assert "WELL" in compsegs
    assert compsegs["WELL"].unique()[0] == "OP1"
    assert len(compsegs.dropna(axis=1, how="all").iloc[0]) == 8

    # Check date handling
    assert "DATE" in compdat_df
    assert not all(compdat_df["DATE"].notna())
    compdat_date = compdat.deck2dfs(deck, start_date="2000-01-01")["COMPDAT"]
    assert "DATE" in compdat_date
    assert all(compdat_date["DATE"].notna())
    assert len(compdat_date["DATE"].unique()) == 1
    assert str(compdat_date["DATE"].unique()[0]) == "2000-01-01"


def test_tstep():
    """Test with TSTEP present"""
    schstr = """
DATES
   1 MAY 2001 /
/

COMPDAT
 'OP1' 33 110 31 31 'OPEN'  /
/

TSTEP
  1 /

COMPDAT
 'OP1' 34 111 32 32 'OPEN' /
/

TSTEP
  2 3 /

COMPDAT
  'OP1' 35 111 33 33 'SHUT' /
/
"""
    deck = EclFiles.str2deck(schstr)
    compdf = compdat.deck2dfs(deck)["COMPDAT"]
    dates = [str(x) for x in compdf["DATE"].unique()]
    assert len(dates) == 3
    assert "2001-05-01" in dates
    assert "2001-05-02" in dates
    assert "2001-05-07" in dates

    schstr_nodate = """
COMPDAT
 'OP1' 33 110 31 31 'OPEN'  /
/

TSTEP
  1 /

COMPDAT
 'OP1' 34 111 32 32 'OPEN' /
/
    """
    assert compdat.deck2dfs(EclFiles.str2deck(schstr_nodate)) == {}
    # (critical error logged)


def test_unrollcompdatk1k2():
    """Test unrolling of k1-k2 ranges in COMPDAT"""
    schstr = """
COMPDAT
  -- K1 to K2 is a range of 11 layers, should be automatically
  -- unrolled to 11 rows.
  'OP1' 33 44 10 20  /
/
"""
    df = compdat.deck2dfs(EclFiles.str2deck(schstr))["COMPDAT"]
    assert df["I"].unique() == 33
    assert df["J"].unique() == 44
    assert (df["K1"].values == range(10, 20 + 1)).all()
    assert (df["K2"].values == range(10, 20 + 1)).all()

    # Check that we can read withoug unrolling:
    df_noroll = compdat.deck2dfs(EclFiles.str2deck(schstr), unroll=False)["COMPDAT"]
    assert len(df_noroll) == 1


def test_samecellperf():
    """Test that we allow multiple perforations in the same cell"""
    schstr = """
COMPDAT
  'OP1' 1 1 1 1 /
  'OP2' 1 1 1 1 /
/
"""
    df = compdat.deck2dfs(EclFiles.str2deck(schstr))["COMPDAT"]
    assert len(df) == 2


def test_unrollwelsegs():
    """Test unrolling of welsegs."""
    schstr = """
WELSEGS
  -- seg_start to seg_end (two first items in second record) is a range of
  -- 2 segments, should be automatically unrolled to 2 rows.
  'OP1' 1689 1923 1.0E-5 'ABS' 'HFA' 'HO' / comment without -- identifier
   2 3 1 1 1923.9 1689.000 0.1172 0.000015  /
/
"""
    df = compdat.deck2dfs(EclFiles.str2deck(schstr))["WELSEGS"]
    assert len(df) == 2

    df = compdat.deck2dfs(EclFiles.str2deck(schstr), unroll=False)["WELSEGS"]
    assert len(df) == 1


def test_unrollbogus():
    """Giving in empty dataframe, should not crash."""
    assert compdat.unrolldf(pd.DataFrame).empty

    bogusdf = pd.DataFrame([0, 1, 4], [0, 2, 5])
    unrolled = compdat.unrolldf(pd.DataFrame([0, 1, 4], [0, 2, 5]), "FOO", "bar")
    # (warning should be issued)
    assert (unrolled == bogusdf).all().all()


def test_initmerging():
    """Test that we can ask for INIT vectors to be merged into the data"""
    eclfiles = EclFiles(REEK)
    noinit_df = compdat.df(eclfiles)
    df = compdat.df(eclfiles, initvectors=[])
    assert isinstance(df, pd.DataFrame)
    assert not df.empty

    df = compdat.df(eclfiles, initvectors=["FIPNUM", "EQLNUM", "SATNUM"])
    assert "FIPNUM" in df
    assert "EQLNUM" in df
    assert "SATNUM" in df
    assert len(df) == len(noinit_df)

    df = compdat.df(eclfiles, initvectors="FIPNUM")
    assert "FIPNUM" in df
    assert len(df) == len(noinit_df)

    with pytest.raises(AssertionError):
        compdat.df(eclfiles, initvectors=2)


def test_main_subparsers(tmp_path, mocker):
    """Test command line interface"""
    tmpcsvfile = tmp_path / "compdat.csv"
    mocker.patch(
        "sys.argv", ["ecl2csv", "compdat", "-v", EIGHTCELLS, "-o", str(tmpcsvfile)]
    )
    ecl2csv.main()

    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert "ZONE" in disk_df
    assert not disk_df.empty

    mocker.patch(
        "sys.argv",
        [
            "ecl2csv",
            "compdat",
            EIGHTCELLS,
            "--initvectors",
            "FIPNUM",
            "-o",
            str(tmpcsvfile),
        ],
    )
    ecl2csv.main()

    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert "FIPNUM" in disk_df
    assert not disk_df.empty

    mocker.patch(
        "sys.argv",
        [
            "ecl2csv",
            "compdat",
            EIGHTCELLS,
            "--initvectors",
            "FIPNUM",
            "EQLNUM",
            "-o",
            str(tmpcsvfile),
        ],
    )
    ecl2csv.main()

    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert "FIPNUM" in disk_df
    assert "EQLNUM" in disk_df
    assert not disk_df.empty


def test_defaulted_compdat_i_j(tmp_path):
    """I and J can be defaulted (that is 1* or 0) in COMPDAT records, then
    that information should be fetched from the most recent WELSPECS keyword
    """

    welspecs_str = """
WELSPECS
  OP1 OPWEST 20 30 1000 /
/
"""
    compdat_str_def_i = """
COMPDAT
  'OP1' 1* 0 10 11  /
/
"""
    compdat_str_def_j = """
COMPDAT
  'OP1' 20 1* 10 11  /
/
"""
    compdat_str_nodefaults = """
COMPDAT
  'OP1' 55 66 80 80  /
/
"""

    with pytest.raises(ValueError, match="WELSPECS must be provided when I"):
        compdat.deck2dfs(EclFiles.str2deck(compdat_str_def_i))["COMPDAT"]

    # I value of 0 also means defaulted:
    with pytest.raises(ValueError, match="WELSPECS must be provided when I"):
        compdat.deck2dfs(EclFiles.str2deck(compdat_str_def_i.replace("1*", "0")))[
            "COMPDAT"
        ]

    with pytest.raises(ValueError, match="WELSPECS must be provided when J"):
        compdat.deck2dfs(EclFiles.str2deck(compdat_str_def_j))["COMPDAT"]

    # J value of 0 also means defaulted:
    with pytest.raises(ValueError, match="WELSPECS must be provided when J"):
        compdat.deck2dfs(EclFiles.str2deck(compdat_str_def_j.replace("1*", "0")))[
            "COMPDAT"
        ]

    with pytest.raises(ValueError, match="WELSPECS must be provided"):
        # Wrong order:
        compdat.deck2dfs(EclFiles.str2deck(compdat_str_def_i + welspecs_str))["COMPDAT"]

    # Simplest example:
    compdat_df = compdat.deck2dfs(EclFiles.str2deck(welspecs_str + compdat_str_def_i))[
        "COMPDAT"
    ]
    assert compdat_df["I"].unique() == [20]
    assert compdat_df["J"].unique() == [30]

    # Two wells:
    compdat_df = compdat.deck2dfs(
        EclFiles.str2deck(
            welspecs_str.replace("OP1", "OP2").replace("30", "99")
            + welspecs_str
            + compdat_str_def_i
        )
    )["COMPDAT"]

    # Partial defaulting
    compdat_df = compdat.deck2dfs(
        EclFiles.str2deck(welspecs_str + compdat_str_def_i + compdat_str_nodefaults)
    )["COMPDAT"]

    assert set(compdat_df["I"].unique()) == {20, 55}
    assert set(compdat_df["J"].unique()) == {30, 66}

    compdat_df = compdat.deck2dfs(
        EclFiles.str2deck(
            welspecs_str.replace("OP1", "OP2").replace("30", "99")
            + welspecs_str
            + compdat_str_def_i
            + compdat_str_def_i.replace("OP1", "OP2")
        )
    )["COMPDAT"]

    assert compdat_df[compdat_df["WELL"] == "OP1"]["I"].unique() == [20]
    assert compdat_df[compdat_df["WELL"] == "OP2"]["I"].unique() == [20]
    assert compdat_df[compdat_df["WELL"] == "OP1"]["J"].unique() == [30]
    assert compdat_df[compdat_df["WELL"] == "OP2"]["J"].unique() == [99]

    # Same well redrilled to new location
    compdat_df = compdat.deck2dfs(
        EclFiles.str2deck(
            "DATES\n  1 JAN 2030 /\n/\n"
            + welspecs_str
            + compdat_str_def_i
            + "DATES\n  1 JAN 2040 /\n/\n"
            + welspecs_str.replace("30", "33")
            + compdat_str_def_i
        )
    )["COMPDAT"]
    assert compdat_df[compdat_df["DATE"].astype(str) == "2030-01-01"]["J"].unique() == [
        30
    ]
    assert compdat_df[compdat_df["DATE"].astype(str) == "2040-01-01"]["J"].unique() == [
        33
    ]


# Multisegement well testing
def test_msw_schfile2df():
    """Test that we can process individual files with AICD and ICD MSW"""
    deck = EclFiles.file2deck(SCHFILE_AICD)
    compdfs = compdat.deck2dfs(deck)
    assert not compdfs["WSEGAICD"].empty
    assert not compdfs["WSEGAICD"].columns.empty

    deck = EclFiles.file2deck(SCHFILE_ICD)
    compdfs = compdat.deck2dfs(deck)
    assert not compdfs["WSEGSICD"].empty
    assert not compdfs["WSEGSICD"].columns.empty

    deck = EclFiles.file2deck(SCHFILE_VALV)
    compdfs = compdat.deck2dfs(deck)
    assert not compdfs["WSEGVALV"].empty
    assert not compdfs["WSEGVALV"].columns.empty


def test_msw_str2df():
    """Testing making a dataframe from an explicit string including MSW"""
    schstr = """
WELSPECS
   'OP_6' 'DUMMY' 28 37 1575.82 OIL 0.0 'STD' 'SHUT' 'YES' 0 'SEG' /
/

COMPDAT
    'OP_6' 28 37 1 1 OPEN 0 1.2719 0.311 114.887 0.0 0.0 'X' 19.65 /
/

WELSEGS
-- WELL   SEGMENTTVD  SEGMENTMD WBVOLUME INFOTYPE PDROPCOMP MPMODEL
   'OP_6'        0.0        0.0   1.0E-5    'ABS'     'HF-'    'HO' /
--  SEG  SEG2  BRANCH  OUT MD       TVD       DIAM ROUGHNESS
     2    2    1        1  2371.596 1577.726  0.15 0.00065    /
/

COMPSEGS
   'OP_6' /
--  I   J   K   BRANCH STARTMD  ENDMD    DIR DEF  SEG
    28  37   1   2     2366.541 2376.651  1*  3*  31   /
/

WSEGAICD
-- WELL SEG SEG2   ALPHA    SF  RHO VIS EMU DEF    X    Y
-- FLAG   A   B   C    D    E    F
   OP_6  31   31 1.7e-05 -1.18 1000 1.0 0.5  4* 3.05 0.67
   OPEN 1.0 1.0 1.0 2.43 1.18 10.0  /
/

WSEGSICD
-- WELL   SEG  SEG2 ALPHA  SF             RHO     VIS  WCT
    OP_6  31   31   0.0001  -1.186915444  1000.0  1.0  0.5  /
/

WSEGVALV
-- WELL   SEG             CV      AC   L
    OP_6  31       0.0084252 0.00075  1*  /
/
"""
    deck = EclFiles.str2deck(schstr)
    compdfs = compdat.deck2dfs(deck)
    wsegaicd = compdfs["WSEGAICD"]
    wsegsicd = compdfs["WSEGSICD"]
    wsegvalv = compdfs["WSEGVALV"]

    # Test WSEGAICD
    assert len(wsegaicd) == 1
    assert "WELL" in wsegaicd
    assert wsegaicd["WELL"].unique()[0] == "OP_6"
    assert len(wsegaicd.dropna(axis=1, how="all").iloc[0]) == 20

    # Test WSEGSICD
    assert len(wsegsicd) == 1
    assert "WELL" in wsegsicd
    assert wsegsicd["WELL"].unique()[0] == "OP_6"
    assert len(wsegsicd.dropna(axis=1, how="all").iloc[0]) == 12

    # Test WSEGVALV
    assert len(wsegvalv) == 1
    assert "WELL" in wsegvalv
    assert wsegvalv["WELL"].unique()[0] == "OP_6"
    assert len(wsegvalv.dropna(axis=1, how="all").iloc[0]) == 5


def test_wsegaicd():
    """Test the WSEGAICD parser for column names and default values"""
    schstr = """WSEGAICD
    OP_6  31   31 1.7e-05 -1.18 1000 1.0 0.5  4* 3.05 0.67
   OPEN 1.0 1.0 1.0 2.43 1.18 10.0  /
/
    """
    deck = EclFiles.str2deck(schstr)
    wsegaicd = compdat.deck2dfs(deck)["WSEGAICD"]
    pd.testing.assert_frame_equal(
        wsegaicd,
        pd.DataFrame(
            data=[
                {
                    "WELL": "OP_6",
                    "SEGMENT1": 31,
                    "SEGMENT2": 31,
                    "STRENGTH": 1.7e-05,
                    "LENGTH": -1.18,
                    "DENSITY_CALI": 1000.0,
                    "VISCOSITY_CALI": 1.0,
                    "CRITICAL_VALUE": 0.5,
                    "WIDTH_TRANS": 0.05,
                    "MAX_VISC_RATIO": 5,
                    "METHOD_SCALING_FACTOR": -1,
                    "MAX_ABS_RATE": None,
                    "FLOW_RATE_EXPONENT": 3.05,
                    "VISC_EXPONENT": 0.67,
                    "STATUS": "OPEN",
                    "OIL_FLOW_FRACTION": 1.0,
                    "WATER_FLOW_FRACTION": 1.0,
                    "GAS_FLOW_FRACTION": 1.0,
                    "OIL_VISC_FRACTION": 2.43,
                    "WATER_VISC_FRACTION": 1.18,
                    "GAS_VISC_FRACTION": 10.0,
                    "DATE": None,
                }
            ]
        ),
    )


def test_wsegsicd():
    """Test the WSEGSICD parser for column names and default values

    Proves bug 232 is fixed.
    """
    schstr = """WSEGSICD
        'WELL_A'              31    31    0.00178 0.57975861     1*   1*        0.7
        1*         1*     1         1*
                         OPEN /
            /
    """
    deck = EclFiles.str2deck(schstr)
    wsegsicd = compdat.deck2dfs(deck)["WSEGSICD"]
    pd.testing.assert_frame_equal(
        wsegsicd,
        pd.DataFrame(
            data=[
                {
                    "WELL": "WELL_A",
                    "SEGMENT1": 31,
                    "SEGMENT2": 31,
                    "STRENGTH": 0.00178,
                    "LENGTH": 0.57975861,
                    "DENSITY_CALI": 1000.25,
                    "VISCOSITY_CALI": 1.45,
                    "CRITICAL_VALUE": 0.7,
                    "WIDTH_TRANS": 0.05,
                    "MAX_VISC_RATIO": 5,
                    "METHOD_SCALING_FACTOR": 1,
                    "MAX_ABS_RATE": None,
                    "STATUS": "OPEN",
                    "DATE": None,
                }
            ]
        ),
    )


def test_wsegvalv():
    """Test the WSEGVALV parser for column names and default values"""
    schstr = """
    WSEGVALV
    -- WELL    SEG         CV      AC  Lex     Dp    ROUGH      Ap  STATUS     A_MAX
    WELL_A   31  0.0084252 0.00075  0.5  0.216   0.0005  0.0366    SHUT    0.0008 /
    /
    """
    deck = EclFiles.str2deck(schstr)
    wsegvalv = compdat.deck2dfs(deck)["WSEGVALV"]
    pd.testing.assert_frame_equal(
        wsegvalv,
        pd.DataFrame(
            data=[
                {
                    "WELL": "WELL_A",
                    "SEGMENT_NUMBER": 31,
                    "CV": 0.0084252,
                    "AREA": 0.00075,
                    "EXTRA_LENGTH": 0.5,
                    "PIPE_D": 0.216,
                    "ROUGHNESS": 0.0005,
                    "PIPE_A": 0.0366,
                    "STATUS": "SHUT",
                    "MAX_A": 0.0008,
                    "DATE": None,
                }
            ]
        ),
    )


def test_wsegvalv_max_blank():
    """Test the WSEGVALV parser for column names and blank values. NB: Blank
    values are taken from other keywords in Eclipse except STATUS, which is OPEN
    by default. Combination of keywords is not tested here."""
    schstr = """
    WSEGVALV
    -- WELL    SEG         CV      AC
    WELL_A   31  0.0084252 0.00075  /
    /
    """
    deck = EclFiles.str2deck(schstr)
    wsegvalv = compdat.deck2dfs(deck)["WSEGVALV"]
    pd.testing.assert_frame_equal(
        wsegvalv,
        pd.DataFrame(
            data=[
                {
                    "WELL": "WELL_A",
                    "SEGMENT_NUMBER": 31,
                    "CV": 0.0084252,
                    "AREA": 0.00075,
                    "EXTRA_LENGTH": None,
                    "PIPE_D": None,
                    "ROUGHNESS": None,
                    "PIPE_A": None,
                    "STATUS": "OPEN",
                    "MAX_A": None,
                    "DATE": None,
                }
            ]
        ),
    )


def test_wsegvalv_max_default():
    """Test the WSEGVALV parser for column names and default values. NB: Default
    values are taken from other keywords in Eclipse except STATUS, which is OPEN
    by default. Combination of keywords is not tested here."""
    schstr = """
    WSEGVALV
    -- WELL    SEG         CV      AC
    WELL_A   31  0.0084252 0.00075  6* /
    /
    """
    deck = EclFiles.str2deck(schstr)
    wsegvalv = compdat.deck2dfs(deck)["WSEGVALV"]
    pd.testing.assert_frame_equal(
        wsegvalv,
        pd.DataFrame(
            data=[
                {
                    "WELL": "WELL_A",
                    "SEGMENT_NUMBER": 31,
                    "CV": 0.0084252,
                    "AREA": 0.00075,
                    "EXTRA_LENGTH": None,
                    "PIPE_D": None,
                    "ROUGHNESS": None,
                    "PIPE_A": None,
                    "STATUS": "OPEN",
                    "MAX_A": None,
                    "DATE": None,
                }
            ]
        ),
    )



"""Test module for layer mapping to zone names"""

from pathlib import Path

import pandas as pd
import pytest

import ecl2df

TESTDIR = Path(__file__).absolute().parent
REEK = str(TESTDIR / "data/reek/eclipse/model/2_R001_REEK-0.DATA")


def test_stdzoneslyr():
    """Test that we can read zones if the zonemap is in a standard location.

    The eclfiles object defines what is the standard location for the file, while
    the actual parsing is done in ecl2df.common.parse_lyrfile() and
    converted to zonemap in common.convert_lyrlist_to_zonemap()
    """
    eclfiles = ecl2df.EclFiles(REEK)

    zonemap = eclfiles.get_zonemap()
    assert isinstance(zonemap, dict)
    assert zonemap[3] == "UpperReek"
    assert zonemap[10] == "MidReek"
    assert zonemap[11] == "LowerReek"
    with pytest.raises(KeyError):
        assert zonemap[0]
    with pytest.raises(KeyError):
        assert zonemap["foo"]
    with pytest.raises(KeyError):
        assert zonemap[-10]
    assert len(zonemap) == 15


def test_nonexistingzones():
    """Test an Eclipse case with non-existing zonemap (i.e. no zonemap file
    in the standard location)"""
    eclfiles = ecl2df.EclFiles(REEK)
    zonemap = eclfiles.get_zonemap("foobar")
    # (we got a warning and an empty dict)
    assert not zonemap


def test_errors(tmp_path, caplog):
    """Test in lyr parse function return correct errors"""
    lyrfile = tmp_path / "formations.lyr"
    lyrfile.write_text(
        """
foo
""",
        encoding="utf-8",
    )
    assert ecl2df.common.parse_lyrfile(lyrfile) is None
    assert "Could not parse lyr file" in caplog.text
    assert "Failed on content: foo" in caplog.text

    lyrfile = tmp_path / "formations.lyr"
    lyrfile.write_text(
        """
valid 1-2
foo 1 2 3
""",
        encoding="utf-8",
    )
    assert ecl2df.common.parse_lyrfile(lyrfile) is None
    assert "Failed on content: foo 1 2 3" in caplog.text

    lyrfile = tmp_path / "formations.lyr"
    lyrfile.write_text(
        """
foo 2-1
""",
        encoding="utf-8",
    )
    assert ecl2df.EclFiles(REEK).get_zonemap(str(lyrfile)) is None
    assert "From_layer higher than to_layer" in caplog.text

    lyrfile = tmp_path / "formations.lyr"
    lyrfile.write_text(
        """
valid 1-2 #FFE5F7
foo   3- 4 #FFGGHH
""",
        encoding="utf-8",
    )
    assert ecl2df.EclFiles(REEK).get_zonemap(str(lyrfile)) is None
    assert "Failed on content: foo   3- 4 #FFGGHH" in caplog.text

    lyrfile = tmp_path / "formations.lyr"
    lyrfile.write_text(
        """
valid 1-2 #FFE5F7
foo   3- 4 bluez
""",
        encoding="utf-8",
    )
    assert ecl2df.EclFiles(REEK).get_zonemap(str(lyrfile)) is None
    assert "Failed on content: foo   3- 4 bluez" in caplog.text

    lyrfile.write_text(
        """
invalid 1-2-3
""",
        encoding="utf-8",
    )
    assert ecl2df.EclFiles(REEK).get_zonemap(str(lyrfile)) is None


def test_lyrlist_format(tmp_path):
    """Ensure the lyr file is parsed correctly"""
    lyrfile = tmp_path / "formations.lyr"
    lyrfile.write_text(
        """
-- Some text
'ZoneA'          1 -     5  #FFE5F7
'ZoneB'          6-     10  --no color
'ZoneC'          11-15    blue
'ZoneD'         3          #fbb
'ZoneE'         19     -20
'ZoneF'         21-22  CORNFLOWERBLUE
""",
        encoding="utf-8",
    )
    lyrlist = ecl2df.common.parse_lyrfile(lyrfile)

    assert lyrlist == [
        {"name": "ZoneA", "from_layer": 1, "to_layer": 5, "color": "#FFE5F7"},
        {
            "name": "ZoneB",
            "from_layer": 6,
            "to_layer": 10,
        },
        {"name": "ZoneC", "from_layer": 11, "to_layer": 15, "color": "blue"},
        {"name": "ZoneD", "span": 3, "color": "#fbb"},
        {
            "name": "ZoneE",
            "from_layer": 19,
            "to_layer": 20,
        },
        {"name": "ZoneF", "from_layer": 21, "to_layer": 22, "color": "CORNFLOWERBLUE"},
    ]


def test_convert_lyrlist_to_zonemap(tmp_path):
    """Test common.covert_lyrlist_to_zonemap()"""
    lyrfile = tmp_path / "formations.lyr"
    lyrfile.write_text(
        """
-- Some text
'ZoneA'          1 -     5
'ZoneB'         5
'ZoneC'         11-20
""",
        encoding="utf-8",
    )
    lyrlist = ecl2df.common.parse_lyrfile(lyrfile)
    zonemap = ecl2df.common.convert_lyrlist_to_zonemap(lyrlist)
    assert zonemap
    assert len(lyrlist) == 3
    assert len(zonemap) == 20
    assert zonemap[10] == "ZoneB"
    assert zonemap[20] == "ZoneC"


def test_nonstandardzones(tmp_path):
    """Test that we can read zones from a specific filename"""
    lyrfile = tmp_path / "formations.lyr"
    lyrfilecontent = """
-- foo
# foo
'Eiriksson'  1-10
 Raude    20-30

# Difficult quote parsing above, might not run in ResInsight.
"""
    lyrfile.write_text(lyrfilecontent)
    lyrlist = ecl2df.common.parse_lyrfile(lyrfile)
    zonemap = ecl2df.common.convert_lyrlist_to_zonemap(lyrlist)
    assert 0 not in zonemap
    assert zonemap[1] == "Eiriksson"
    assert zonemap[10] == "Eiriksson"
    assert 11 not in zonemap
    assert 19 not in zonemap
    assert zonemap[20] == "Raude"
    assert zonemap[30] == "Raude"
    assert len(zonemap) == 21


@pytest.mark.parametrize(
    "dframe, zonedict, zoneheader, kname, expected_df",
    [
        (
            pd.DataFrame([{}]),
            {},
            "ZONE",
            "K",
            pd.DataFrame([{}]),
        ),
        (
            pd.DataFrame([{"K": 1}]),
            {},
            "ZONE",
            "K",
            pd.DataFrame([{"K": 1}]),
        ),
        (
            pd.DataFrame([{"K": 1}]),
            {1: "FOO"},
            "ZONE",
            "K",
            pd.DataFrame([{"K": 1, "ZONE": "FOO"}]),
        ),
        (
            pd.DataFrame([{"KK": 1}]),
            {1: "FOO"},
            "ZONE",
            "KK",
            pd.DataFrame([{"KK": 1, "ZONE": "FOO"}]),
        ),
        (
            pd.DataFrame([{"K": 1}]),
            {1: "FOO"},
            "ZONE",
            "KK",
            pd.DataFrame([{"K": 1}]),
        ),
        (
            pd.DataFrame([{"K": 1}]),
            {1: "FOO"},
            "SUBZONE",
            "K",
            pd.DataFrame([{"K": 1, "SUBZONE": "FOO"}]),
        ),
        (
            pd.DataFrame([{"K": 1}]),
            {2: "FOO"},
            "ZONE",
            "K",
            pd.DataFrame([{"K": 1, "ZONE": None}]),
        ),
        (
            pd.DataFrame([{"K": 1}, {"K": 2}]),
            {2: "FOO"},
            "ZONE",
            "K",
            pd.DataFrame([{"K": 1, "ZONE": None}, {"K": 2, "ZONE": "FOO"}]),
        ),
    ],
)
def test_merge_zones(dframe, zonedict, zoneheader, kname, expected_df):
    pd.testing.assert_frame_equal(
        ecl2df.common.merge_zones(dframe, zonedict, zoneheader, kname),
        expected_df,
        check_like=True,
    )


def test_repeated_merge_zone():
    """Merging zone information into a frame where the column already
    exists should not modify the frame."""

    dframe = pd.DataFrame([{"K1": 1, "ZONE": "upper"}])
    pd.testing.assert_frame_equal(
        ecl2df.common.merge_zones(dframe, {1: "upper"}, "ZONE"), dframe
    )
    pd.testing.assert_frame_equal(
        ecl2df.common.merge_zones(dframe, {1: "lower"}, "ZONE"), dframe
    )
